#  ---------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System -
#  See license.txt file in the root directory of this Python module source  
#  ---------------------------------------------------------------------------
#
from __future__ import unicode_literals
from builtins import (str, bytes, range, dict)

TAG = "\n### pyCGNS:"
TAG_ERROR = " ERROR "


def perr(error_id, *tp):
    try:
        msg = TAG + TAG_ERROR + "[%.3d] %s" % (error_id, errorTable[error_id])
    except (TypeError, KeyError):
        msg = TAG + TAG_ERROR + "[%.3d] %s" % (error_id, errorTable[999])
    ret = msg
    if tp:
        if isinstance(tp, tuple) and (len(tp) > 0):
            ret = msg % tp[0]
        else:
            ret = msg % tp
    return ret


class CGNSException(Exception):
    def __init__(self, code, args=None):
        self.code = code
        self.value = args

    def __str__(self):
        if self.value is not None:
            ret = perr(self.code, self.value)
        else:
            ret = perr(self.code)
        return ret

cgnsException = CGNSException
class cgnsNodeError(cgnsException):
    pass


class cgnsNameError(cgnsNodeError):
    pass


class cgnsTypeError(cgnsNodeError):
    pass


# -----------------------------------------------------------------------------
errorTable = {
    -1: "Stop on check",
    0: "No error",
    1: "Node is empty !",
    2: "Node should be a list of <name, value, children, type>",
    3: "Node name should be a string",
    4: "Node [%s] children list should be a list",
    5: "Node [%s] bad value: should be a numpy object",
    6: "Parent node for [%s] is not a CGNS node",
    7: "Failed to add a badly formed child node into [%s]",
    8: "Failed to add root node as child node into [%s]",
    9: "Link chaser returns a badly formed node as place of [%s]",
    10: "CGNSBase [%s] bad cell dimensions",
    11: "CGNSBase [%s] bad physical dimensions",
    12: "CGNSBase [%s] bad cell/physical dimensions values",
    20: "No node of type [%s] with name [%s]",
    21: "No node with name [%s]",
    22: "Node name should have type string",
    23: "Empty string is not allowed for a node name",
    24: "Node name should not contain a '/' (you have [%s])",
    25: "Node name length should not be greater than 32 chars: %s",
    26: "Bad name [%s] for this node",
    27: "Bad type [%s] for this node",
    28: "Node [%s] should have no child",
    29: "Node name cannot be '.' or '..'",
    30: "Nodes [%s] and [%s] are different",
    31: "Node name cannot be only ' '",
    32: "Node name should not have prefix or suffix ' 's",
    33: "Node name contains unsafe chars",
    34: "Node name contains two consecutive spaces",
    40: "Node [%s] has type [%s] which is not a CGNS/SIDS type",
    90: "Node is None!",
    91: "Root node can have only CGNSBase_t or CGNSLibraryVersion_t children",
    99: "More than one CGNSLibraryVersion node found",
    100: "Absolute path implies a root node with CGNSLibraryVersion node",
    101: "No child on CGNS tree first level",
    102: "Duplicated child name [%s] in [%s]",
    103: "Node type of [%s] not [%s]",
    104: "Node type of [%s] not in %s",
    105: "String data should have type array or string",
    106: "Value should be a real array",
    107: "Value should be an integer array",
    108: "Parent node should have type [%s]",
    109: "Value should be an array",
    110: "Value should be a string in [%s]",
    111: "Bad value type",
    112: "Empty args is not allowed as value",
    200: "Bad GridLocation value [%s]",
    201: "Bad ConvergenceHistory name [%s]",
    202: "Units specification list should have 5 values (one per unit)",
    203: "Bad units specification [%s]",
    204: "Units specifications duplicated [%s]",
    205: "Bad SimulationType value [%s]",
    206: "Bad ZoneType value [%s]",
    207: "Bad DataClass value [%s]",
    208: "Exponents specification list should have 5 values (one per unit)",
    209: "BaseIterativeData number of steps should be an integer value",
    210: "BaseIterativeData steps should in %s",
    211: "BaseIterativeData value has greater length than number of steps",
    212: "BaseIterativeData is missing sub-node",
    213: "Parent node should be BaseIterativeData",
    214: "BaseIterativeData bad Pointer name [%s]",
    215: "BaseIterativeData Pointer list [%s] has not the right length [%s]",
    216: "BaseIterativeData Pointer list [%s] contents has bad length [%s]",
    217: "Parent node should be EMConductivityModel",
    218: "Bad EMConductivityModelType value [%s]",
    219: "EMConductivityModelType is missing sub-node",
    220: "Parent node should be GoverningEquations",
    221: "Bad GoverningEquationsType value [%s]",
    222: "GoverningEquationsType is missing sub-node",
    223: "Parent node should be GasModel",
    224: "Bad GasModelType value [%s]",
    225: "GasModelType is missing sub-node",
    226: "Parent node should be ThermalConductivityModel",
    227: "Bad ThermalConductivityModelType value [%s]",
    228: "ThermalConductivityModelType is missing sub-node",
    229: "Parent node should be ViscosityModel",
    230: "Bad ViscosityModelType value [%s]",
    231: "ViscosityModelType is missing sub-node",
    232: "Parent node should be TurbulenceClosure",
    233: "Bad TurbulenceClosureType value [%s]",
    234: "TurbulenceClosureType is missing sub-node",
    235: "Parent node should be TurbulenceModel",
    236: "Bad TurbulenceModelType value [%s]",
    237: "TurbulenceModelType is missing sub-node",
    238: "Parent node should be ThermalRelaxationModel",
    239: "Bad ThermalRelaxationModelType value [%s]",
    240: "ThermalRelaxationModelType is missing sub-node",
    241: "Parent node should be ChemicalKineticsModel",
    242: "Bad ChemicalKineticsModelType value [%s]",
    243: "ChemicalKineticsModelType is missing sub-node",
    244: "Parent node should be EMElectricFieldModel",
    245: "Bad EMElectricFieldModelType value [%s]",
    246: "EMElectricFieldModelType is missing sub-node",
    247: "Parent node should be EMMagneticFieldModel",
    248: "Bad EMMagneticFieldModelType value [%s]",
    249: "EMMagneticFieldModelType is missing sub-node",
    250: "Bad ElementType value [%s]",
    251: "Bad ElementConnectivity value [%s]",
    252: "Bad BCTypeSimple value [%s]",
    253: "Bad AverageInterfaceType value [%s]",
    254: "Bad RigidGridMotionType value [%s]",
    255: "Bad ArbitraryGridMotionType value [%s]",
    256: "Bad GeometryFormat value [%s]",
    257: "Bad BCTypeSimple or BCTypeCompound value [%s]",
    258: "Parent node should be ArbitraryGridMotion_t",
    259: "Parent node should be ReferenceState_t",
    260: "Parent node should be IntegralData_t",
    261: "Parent node should be UserDefinedData_t",
    262: "Parent node should be Family_t",
    263: "Parent node should be GridCoordinates_t",
    264: "Parent node should be DiscreteData_t",
    265: "Parent node should be BCData_t",
    266: "Parent node should be ZoneIterativeData_t",
    267: "Parent node should be AverageInterface_t",
    268: "Bad AverageInterfaceType value [%s]",
    269: "AverageInterfaceType is missing sub-node",
    270: "Parent node should be ZoneGridConnectivity_t",
    271: "Bad ZoneGridConnectivityType value [%s]",
    272: "ZoneGridConnectivityType is missing sub-node",
    273: "Parent node should be DimensionalExponents_t",
    274: "Bad SimulationType value [%s]",
    275: "Parent node should be BCDataSet_t",
    276: "BCDataSetType is missing sub-node",
    277: "Parent node should be ZoneBC_t",
    278: "Parent node should be GridCoordinates_t",
    279: "Parent node should be RotatingCoordinates_t",
    280: "Parent node should be Axisymmetry_t",
    281: "Parent node should be FlowSolution_t ",
    282: "Parent node should be Periodic_s ",
    283: "Parent node should be OversetHoles_t ",
    284: "Parent node should be GridConnectivity1to1_t ",
    285: "Parent node should be GridCoordinates",
    286: "Parent node should be DiscreteData_t",
    287: "Parent node should be DataArray_t",
    288: "Parent node should be Zone_t",
    289: "Parent node should be DataConversion_t",
    290: "Parent node should be GridLocation_t",
    291: "Parent node should be Descriptor_t",
    292: "Parent node should be GeometryReference_t",
    293: "GeometryReferenceType is missing sub-node",
    294: "Parent node should be FamilyBC_t",
    295: "FamilyBCType is missing sub-node",

    300: "Zone creation requires a dimension tuple",

    710: "Numpy array should have a 'Fortran' order",

    800: "Cannot find pyCGNS config ?",
    801: "No information about numeric library in pyCGNS config",
    802: "Bad information about numeric library in pyCGNS config [%s]",

    900: 'No such file [{}]',
    901: 'File already exists [{}]',
    902: 'File name should be a string',
    903: 'Cannot propagate save to non-existant linked-to file [{}]',
    904: 'Cannot propagate save to unreadable linked-to file [{}]',
    905: 'Cannot propagate save to non-existant node in linked-to file [{}]',
    906: 'Bad tree structure (root should be list of 4 items, refer to doc)',
    907: 'Invalid combination of flags (refer to doc)',
    908: 'Using a maxdata implies FNODATA flag',
    909: 'Cannot have attributes {} at the same time',
    910: 'Keyword argument [{}] unknown',
    911: 'Bad [flag] argument (should be int)',
    912: 'Bad [depth] argument (should be int)',
    913: 'Bad [path] argument (should be str)',
    914: 'Bad [lksearch] argument (should be list of str)',
    915: 'Bad [maxdata] argument (should be int)',
    916: 'Bad [skip] argument (should be list of str)',
    917: 'Bad [links] argument (refer to doc)',
    920: 'Cannot write in file [{}]',
    930: 'Filter dict requires a path (str) as key',
    931: 'Filter dict value should be a tuple with int as first arg',
    932: 'SPARTIAL Filter requires a "list of 8 lists of integers" as value"',
    933: 'SPARTIAL Filter bad parameters',
    934: 'SCONTIGUOUS requires a tuple of index,rank (integers) as value',

    999: "Unknow error code",
}

CGNS_NoSuchFile = 'pyCGNS[001] No such file'
CGNS_OpenFileFailed = 'pyCGNS[002] Open file failed'
CGNS_FileAlreadyExists = 'pyCGNS[003] File already exists'
CGNS_NotConnected = "pyCGNS[100] Not yet connected to a CGNS database"
CGNS_NoSuchBase = "pyCGNS[101] No such base name"
CGNS_AbsolutePathRequired = "pyCGNS[102] Absolute path required"
CGNS_NoSuchZone = "pyCGNS[103] No such zone name"
CGNS_NoSuchArray = "pyCGNS[104] No such array name"
CGNS_NoSuchUserData = "pyCGNS[105] No such user data node name"
CGNS_NoSuchFlowSolution = "pyCGNS[106] No such flow solution node name"
CGNS_NoSuchBoundary = "pyCGNS[107] No such boundary condition node name"
CGNS_FileShouldBeReadable = "pyCGNS[110] File should be readable"
CGNS_GotoPathNotFound = "pyCGNS[901] No such goto path in CGNS tree"
CGNS_BadType = "pyCGNS[200] Bad type"
CGNS_BadOpenMode = "pyCGNS[201] Bad OpenMode value"
CGNS_BadErrorCode = "pyCGNS[202] Bad ErrorCode type"
CGNS_BadMassUnit = "pyCGNS[203] Bad MassUnit type"
CGNS_BadLengthUnit = "pyCGNS[204] Bad LengthUnit type"
CGNS_BadTimeUnit = "pyCGNS[205] Bad TimeUnit type"
CGNS_BadTemperatureUnit = "pyCGNS[206] Bad TemperatureUnit type"
CGNS_BadAngleUnit = "pyCGNS[207] Bad AngleUnit type"
CGNS_BadDataClass = "pyCGNS[208] Bad DataClass type"
CGNS_BadGridLocation = "pyCGNS[209] Bad GridLocation type"
CGNS_BadBCDataType = "pyCGNS[210] Bad BC data type"
CGNS_BadGridConnectivityType = "pyCGNS[211] Bad GridConnectivityType type"
CGNS_BadPointSetType = "pyCGNS[212] Bad PointSetType type"
CGNS_BadGoverningType = "pyCGNS[213] Bad GoverningEquation type"
CGNS_BadModelType = "pyCGNS[214] Bad Model type"
CGNS_BadBCType = "pyCGNS[215] Bad BC type"
CGNS_BadDataType = "pyCGNS[216] Bad DataType type"
CGNS_BadElementType = "pyCGNS[217] Bad ElementType type"
CGNS_BadZoneType = "pyCGNS[218] Bad ZoneType type"
CGNS_BadRigidGridMotionType = "pyCGNS[219] Bad RigidGridMotionType type"
CGNS_BadArbitraryGridMotionType = "pyCGNS[220] Bad ArbitraryGridMotionType type"
CGNS_BadSimulationType = "pyCGNS[221] Bad SimulationType type"
CGNS_BadWallFunctionType = "pyCGNS[222] Bad WallFunctionType type"
CGNS_BadAreaType = "pyCGNS[223] Bad AreaType type"
CGNS_BadAverageInterfaceType = "pyCGNS[224] Bad AverageInterfaceType type"

# -----------------------------------------------------------------------------



#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System - 
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
import numpy as NPY


def transformCheckValues(data, cdim):
    if (int(cdim) not in (1, 2, 3)): return False
    adata = map(abs, data)
    if ((cdim == 1) and (1 not in adata)): return False
    if ((cdim == 2) and (set(adata) != set((1, 2)))): return False
    if ((cdim == 3) and (set(adata) != set((1, 2, 3)))): return False
    return True


def transformAsVector(data, cdim):
    v = [None] * cdim
    for n in range(cdim):
        v[n] = [0, 0, 0]
        v[n][abs(data[n]) - 1] = 1 * NPY.sign(data[n])
    return v


def transformIsRightHanded(data, cdim):
    return transformIsDirect(data, cdim)


def transformIsDirect(data, cdim):
    if (not transformCheckValues(data, cdim)):
        return False
    v = transformAsVector(data, cdim)
    if cdim == 3:
        return NPY.all(v[2] == NPY.cross(v[0], v[1]))
    return True



#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System -
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
from __future__ import unicode_literals
import numpy
import h5py
import os

from ..PAT import cgnskeywords as CGK
from ..PAT import cgnsutils as CGU
from ..PAT import cgnserrors as CGE

class Flags(object):
    pass

flags = Flags()
flags.NONE = 0x00000000
flags.ALL = 0xFFFFFFFF
flags.TRACE = 0x00000001
flags.FOLLOWLINKS = 0x00000002
flags.NODATA = 0x00000004
flags.KEEPLIST = 0x00000008
flags.MERGELINKS = 0
flags.COMPRESS = 0x00000010
flags.REVERSEDIMS = 0x00000020
flags.OWNDATA = 0x00000040
flags.UPDATE = 0x00000080
flags.DEBUG = 0x00008000
flags.DELETEMISSING = 0x00000100
flags.ALTERNATESIDS = 0x00000200
flags.NOTRANSPOSE = 0
flags.UPDATEONLY = 0x00000400
flags.FORTRANFLAG = 0x00000800
flags.PROPAGATE = 0x00004000
flags.LINKOVERRIDE = 0x00010000
flags.DEFAULT = flags.FOLLOWLINKS\
              | flags.DELETEMISSING\
              | flags.OWNDATA\
              | flags.REVERSEDIMS\
              | flags.FORTRANFLAG\
              | flags.OWNDATA\
              | flags.LINKOVERRIDE
flags.DEFAULTS = flags.DEFAULT
flags.CHECKSUM = 0x00080000

flags.links=Flags()
flags.links.OK = 0x0000
flags.links.FAIL = 0x0001
flags.links.BADSYNTAX = 0x0002
flags.links.NOFILE = 0x0004
flags.links.FILENOREAD = 0x0008
flags.links.NONODE = 0x0010
flags.links.LOOP = 0x0020
flags.links.IGNORED = 0x0040
flags.links.UPDATED = 0x0080

def raiseException(code, *args):
    if args:
        raise MAPException((code, MAPException.mTable[code].format(args)))
    else:
        raise MAPException((code, MAPException.mTable[code]))

def probe(filename, path=None):
    tfile = os.path.normpath(os.path.expanduser(filename))
    if not os.path.isfile(tfile):
        raiseException(900, tfile)
    return True

def load(filename, nodata = False, maxdata = 65, subtree = '', follow_links = False, **kwargs):
    """
    Load a CGNS/HDF5 CGNS tree as a CGNS/Python object::

      from CGNS.MAP import load
      
      (tree, links, paths) = load('cgnsfile.cgns')

    Detailed documentation about the load can be found 
    `here <https://pycgns.github.io/MAP/_index.html>`_, the following doc only
    describes the function signature.

    :arg str filename: CGNS/HDF5 file to open
    :arg bool nodata: if `True` the actual node data is load only if below the `maxdata`
    :arg int maxdata: if `nodata` is `True`, `maxdata` gives the threshold for loading actual data
    :arg str subtree: loads only starting from the `subtree` path arg
    :arg bool follow_links: If `True` follows the linked-to files (default:`False`)

    :return: a tuple (tree, links, paths) see remaks below

    :Remarks:
      - `tree` is a Python list, the actual CGNS/Python tree created from the CGNS/HDF5 file.
      - `links` is the list of linked-to files and nodes used during the `tree` creation.
      - `paths` is the list of nodes paths in `tree` you have to take care of such as no-data nodes.

    """
    def _load(grp,links,paths,load_children = True):
        
        if grp.attrs['type'] == 'LK':
            if not follow_links:
                links.append(['',
                              grp[' file'].value[:-1].tostring().decode('ascii'),
                              grp[' path'].value[:-1].tostring().decode('ascii'),
                              grp.name,
                              64])
                return None
            else:
                grp = grp[' link']
            
        node = [unicode(grp.attrs['name']),
                None,
                [],
                unicode(grp.attrs['label'])
                ]
        
        if ' data' in grp:
            dset  = grp[' data']
            n = dset.size 
            if not nodata or n < maxdata:
                load_data = True
            else:
                load_data = False
                paths.append([grp.name,
                              1,
                              grp.attrs['type'],
                              dset.shape
                              ])
            
            if load_data:
                node[1] = numpy.copy(dset.value.transpose(),order = 'F')
                if grp.attrs['type'] == 'C1':
                    node[1] = numpy.vectorize(chr)(node[1]).astype(numpy.dtype('c'))
                
        if load_children:
            cnames = [x for x in grp.keys() if x[0] != ' ']
            for cname in cnames:
                child = _load(grp[cname],links,paths)
                if child is not None:
                    node[2].append(child)
            
        return node
        
    links = []
    paths = []
    with h5py.File(filename,'r') as h5f:
        if subtree == '':
            tree = _load(h5f,links,paths)
        else:
            grp = h5f[subtree]
            node = _load(grp,links,paths)
            while grp.name != '/':
                pnode = _load(grp.parent,links,paths,load_children = False)
                pnode[2].append(node)
                node = pnode
                grp = grp.parent

            tree = node
        
    tree[0] = u'CGNSTree'
    tree[3] = u'CGNSTree_t'

    return tree,links,paths

def save(filename,tree, links = [], **kwargs):
    
    def _cst_size_str(s,n):
            
        return numpy.string_(s.ljust(n,'\x00'))
        
    def _save(grp,node):
        
        _cst_size_str(node[0],33)
        
        if node[3] == 'CGNSTree_t':
            grp.attrs['name']  = _cst_size_str('HDF5 MotherNode', 33)
            grp.attrs['label'] = _cst_size_str('Root Node of HDF5 File', 33)
        else:
            grp.attrs['name']  = _cst_size_str(node[0], 33)
            grp.attrs['label'] = _cst_size_str(node[3], 33)
            grp.attrs['flags'] = numpy.array([0],dtype = numpy.int32)
            
        if node[1] is None:
            grp.attrs['type'] = _cst_size_str("MT", 3)
        else:
            if node[1].dtype == numpy.float32:
                grp.attrs['type'] = _cst_size_str("R4", 3)
            elif node[1].dtype == numpy.float64:
                grp.attrs['type'] = _cst_size_str("R8", 3)
            elif node[1].dtype == numpy.int32:
                grp.attrs['type'] = _cst_size_str("I4", 3)
            elif node[1].dtype == numpy.int64:
                grp.attrs['type'] = _cst_size_str("I8", 3)
            elif node[1].dtype == numpy.dtype('|S1'):
                grp.attrs['type'] = _cst_size_str("C1", 3)
            else:
                raise NotImplementedError("unknown dtype %s"%node[1].dtype)
        
            data = node[1]
            if node[1].dtype == numpy.dtype('|S1'):
                idx1 = numpy.nonzero(data != '')
                idx2 = numpy.nonzero(data == '')
                tmp = data.copy()
                data = numpy.zeros(data.shape,dtype = numpy.int8)
                data[idx1] = numpy.vectorize(ord)(tmp[idx1])
                data[idx2] = 0
                data = data.astype(numpy.int8)
            
            data = data.transpose()
            grp.create_dataset(' data',data = data)
            
        
        for child in node[2]:
            sgrp = grp.create_group(child[0])
            _save(sgrp,child)
            
    def _add_links(h5f,links):
        
        for dname,filename,rpth,lpth,n in links:
            tmp = lpth.split('/')
            name = tmp[-1]
            ppth = '/'.join(tmp[:-1])
            
            grp = h5f[ppth]
            sgrp = grp.create_group(name)
            
            sgrp.attrs['name']  = _cst_size_str(name, 33)
            sgrp.attrs['label'] = _cst_size_str('', 33)
            sgrp.attrs['type'] = _cst_size_str("LK", 3)
            
            data = numpy.array([ord(x) for x in rpth+'\x00'],dtype = numpy.int8)
            sgrp.create_dataset(' path',data = data)
            data = numpy.array([ord(x) for x in filename+'\x00'],dtype = numpy.int8)
            sgrp.create_dataset(' file',data = data)
            sgrp[' link'] = h5py.ExternalLink(filename, rpth)
            
            assert dname == ''
            
            
            
    with h5py.File(filename,'w') as h5f:
        h5f.create_dataset(' format', data = numpy.array([78,65,84,73,86,69,0],dtype = numpy.int8))
        h5f.create_dataset(' hdf5version', data = numpy.array([ 72,68,70,53,32,86,101,114,115,105,111,110,32,49,46,56,46,49
,53,0,0,0,0,0,0,0,0,0,0,0,0,0,0],dtype = numpy.int8))
        _save(h5f,tree)
        _add_links(h5f,links)
        
#
# --- last line



#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System -
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
# TESTING NAV (imports & functions, no GUI) ***
#
from __future__ import unicode_literals
from __future__ import print_function
import unittest

import CGNS.PAT.cgnslib as CGL
import CGNS.PAT.cgnsutils as CGU
import CGNS.PAT.cgnskeywords as CGK
import numpy as NPY

import importlib
import os
import string
import subprocess


def genTrees():
    tree = CGL.newCGNSTree()
    b = CGL.newBase(tree, '{Base}', 2, 3)
    z = CGL.newZone(b, '{Zone}', NPY.array([[5, 4, 0], [7, 6, 0]], order='F'))
    g = CGL.newGridCoordinates(z, 'GridCoordinates')
    d = CGL.newDataArray(g, CGK.CoordinateX_s, NPY.ones((5, 7), dtype='d', order='F'))
    d = CGL.newDataArray(g, CGK.CoordinateY_s, NPY.ones((5, 7), dtype='d', order='F'))
    d = CGL.newDataArray(g, CGK.CoordinateZ_s, NPY.ones((5, 7), dtype='d', order='F'))
    return (tree,)


class NAVTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def test_000_Module(self):
        import qtpy
        import qtpy.QtCore
        import qtpy.QtWidgets

    def test_001_Script(self):
        import CGNS.NAV


# ---
print('-' * 70 + '\nCGNS.NAV test suite')
suite = unittest.TestLoader().loadTestsFromTestCase(NAVTestCase)
unittest.TextTestRunner(verbosity=2).run(suite)

# --- last line



#  -------------------------------------------------------------------------
#  pyCGNS.NAV - Python package for CFD General Notation System - NAVigater
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
from __future__ import unicode_literals
from __future__ import print_function
from builtins import (str, bytes, range, dict)

from CGNS.NAV.moption import Q7OptionContext as OCTXT

import sys
import numpy
import os

import CGNS.PAT.cgnsutils as CGU
import CGNS.PAT.cgnskeywords as CGK

from qtpy.QtCore import Qt
from qtpy.QtWidgets import (QFileDialog, QStyledItemDelegate, QLineEdit, QComboBox, QTableWidgetItem, QMessageBox)
from qtpy.QtGui import (QFont, QIcon, QPixmap)

from CGNS.NAV.Q7QueryWindow import Ui_Q7QueryWindow
from CGNS.NAV.Q7SelectionWindow import Ui_Q7SelectionWindow
from CGNS.NAV.wfingerprint import Q7Window
from CGNS.NAV.mquery import Q7QueryEntry
from CGNS.NAV.mtree import (COLUMN_VALUE, COLUMN_DATATYPE,
                            COLUMN_SIDS, COLUMN_NAME,
                            HIDEVALUE)
import CGNS.NAV.wmessages as MSG

(CELLCOMBO, CELLTEXT) = range(2)
CELLEDITMODE = (CELLCOMBO, CELLTEXT)


# -----------------------------------------------------------------
class Q7SelectionItemDelegate(QStyledItemDelegate):
    def __init__(self, owner, model):
        QStyledItemDelegate.__init__(self, owner._tb)
        self._parent = owner
        self._mode = CELLTEXT
        self._model = model
        self._lastCol = None

    def createEditor(self, parent, option, index):
        tnode = self._model.nodeFromPath(self._parent._data[index.row()])
        if tnode.sidsIsCGNSTree():
            return None
        if tnode.sidsIsLink():
            return None
        if tnode.sidsIsLinkChild():
            return None
        ws = option.rect.width()
        hs = option.rect.height() + 4
        xs = option.rect.x()
        ys = option.rect.y() - 2
        if index.column() == 2:
            self._lastCol = COLUMN_SIDS
            if self._parent.cForce.checkState() != Qt.Checked:
              self._mode = CELLTEXT
              editor = QLineEdit(parent)
              editor.transgeometry = (xs, ys, ws, hs)
            else:
              self._mode = CELLCOMBO
              editor = QComboBox(parent)
              editor.transgeometry = (xs, ys, ws, hs)
              itemslist = tnode.sidsTypeList()
              editor.addItems(itemslist)
              try:
                tix = itemslist.index(tnode.sidsType())
              except ValueError:
                editor.insertItem(0, tnode.sidsType())
                tix = 0
              editor.setCurrentIndex(tix)
            editor.installEventFilter(self)
            self.setEditorData(editor, index)
            return editor
        if index.column() == 3:
            self._lastCol = COLUMN_DATATYPE
            self._mode = CELLCOMBO
            editor = QComboBox(parent)
            editor.transgeometry = (xs, ys, ws, hs)
            itemslist = tnode.sidsDataType(all=True)
            editor.addItems(itemslist)
            editor.setCurrentIndex(0)
            editor.installEventFilter(self)
            self.setEditorData(editor, index)
            return editor
        if index.column() == 4:
            self._lastCol = COLUMN_VALUE
            if tnode.hasValueView():
                pt = tnode.sidsPath().split('/')[1:]
                lt = tnode.sidsTypePath()
                fc = self._parent._control.userFunctionFromPath(pt, lt)
                if fc is not None:
                    en = fc.getEnumerate(pt, lt)
                else:
                    en = tnode.sidsValueEnum()
                if en is None:
                    self._mode = CELLTEXT
                    editor = QLineEdit(parent)
                    editor.transgeometry = (xs, ys, ws, hs)
                else:
                    self._mode = CELLCOMBO
                    editor = QComboBox(parent)
                    editor.transgeometry = (xs, ys, ws, hs)
                    editor.addItems(en)
                    try:
                        tix = en.index(tnode.sidsValue().tostring().decode('ascii'))
                    except ValueError:
                        editor.insertItem(0, tnode.sidsValue().tostring().decode('ascii'))
                        tix = 0
                    editor.setCurrentIndex(tix)
                editor.installEventFilter(self)
                self.setEditorData(editor, index)
                return editor
        return None

    def setEditorData(self, editor, index):
        if self._mode == CELLTEXT:
            value = index.data()
            editor.clear()
            editor.insert(value)
        elif self._mode == CELLCOMBO:
            value = index.data()
            ix = editor.findText(value)
            if ix != -1:
                editor.setCurrentIndex(ix)
        else:
            pass

    def setModelData(self, editor, model, index):
        col = self._lastCol
        if col is None:
            return
        value = None
        nodelist = []
        dt = self._parent._data
        if self._mode == CELLCOMBO:
            value = editor.currentText()
        if self._mode == CELLTEXT:
            value = editor.text()
        if self._parent.setToAll():
            reply = MSG.wQuestion(self._parent, 371, 'Apply changes to all...',
                                  """The same value is going to be set to all selected nodes,<br>
                                     Do we continue and perform the modification on all selected?<br>
                                     This may fail on some nodes if the data type is not the same
                                     as the original node for example. In that case some of the
                                     nodes would be modified but not the others, you would not
                                     have any status on the actual result of the modification.<br>
                                     """)
            if reply == MSG.OK:
                for rw in self._parent.selectedRows():
                    nd = self._model.nodeFromPath(dt[rw])
                    nodelist.append((rw, nd))
            else:
                return
        else:
            nodelist.append((index.row(),
                             self._model.nodeFromPath(dt[index.row()])))
        if self._parent.cForce.checkState() != Qt.Checked:
            noforce = True
        else:
            noforce = False
        for ndr in nodelist:
            nix = self._model.createIndex(ndr[0], col, ndr[1])
            if noforce:
                nix.internalPointer().setEditCheck(False)
            self._model.setData(nix, value, role=Qt.EditRole)
            self._parent.updateRowData(ndr[0], dt[ndr[0]])

    def updateEditorGeometry(self, editor, option, index):
        editor.setGeometry(*editor.transgeometry)


# -----------------------------------------------------------------
class Q7SelectionList(Q7Window, Ui_Q7SelectionWindow):
    def __init__(self, parent, model, fgindex):
        Q7Window.__init__(self, Q7Window.VIEW_SELECT,
                          parent._control, None, fgindex)
        self.bClose.clicked.connect(self.reject)
        self._parent = parent
        self._model = model
        self._data = model.getSelected()
        self._tb = self.selectionTable
        self.bSave.clicked.connect(self.selectionsave)
        self.bInfo.clicked.connect(self.infoSelectView)
        self.cShowPath.stateChanged.connect(self.colCheck)
        self.cShowSIDS.stateChanged.connect(self.colCheck)
        # QObject.connect(self.cShowPath,
        #                SIGNAL("stateChanged(int)"),self.colCheck)
        # QObject.connect(self.cShowSIDS,
        #                SIGNAL("stateChanged(int)"),self.colCheck)
        self.bFirst.clicked.connect(self.sClear)
        self._tb.setItemDelegate(Q7SelectionItemDelegate(self, self.FG.model))
        self.bRemoveToSelect.clicked.connect(self.sRemove)
        self.bReverse.clicked.connect(self.sReverse)
        self.bSelectAll.clicked.connect(self.sAll)
        self.bUnselectAll.clicked.connect(self.sClear)

    def doRelease(self):
        pass

    def colCheck(self):
        if self.cShowPath.checkState() == Qt.Checked:
            self._tb.showColumn(0)
        else:
            self._tb.hideColumn(0)
        if self.cShowSIDS.checkState() == Qt.Checked:
            self._tb.showColumn(2)
        else:
            self._tb.hideColumn(2)

    def setToAll(self):
        if self.cApplyToAll.checkState() == Qt.Checked:
            return True
        return False

    def selectedRows(self):
        ixl = self._tb.selectionModel().selection().indexes()
        sr = set()
        for ix in ixl:
            sr.add(ix.row())
        return list(sr)

    def sRemove(self):
        dsel = []
        for r in self._tb.selectionModel().selectedRows():
            dsel.append(self._data[r.row()])
        for p in dsel:
            self._data.remove(p)
        self.reset()

    def sReverse(self):
        rall = set(range(self._tb.rowCount()))
        rsel = set()
        for r in self._tb.selectionModel().selectedRows():
            rsel.add(r.row())
        runs = rall.difference(rsel)
        self._tb.clearSelection()
        for r in runs:
            self._tb.selectRow(r)

    def sAll(self):
        self._tb.selectAll()

    def sClear(self):
        self._tb.clearSelection()

    def infoSelectView(self):
        self._control.helpWindow('Selection')

    def show(self):
        self.reset()
        super(Q7SelectionList, self).show()
        self.raise_()

    def selectionsave(self):
        n = 'data=[\n'
        for path in self._data:
            node = self._model.nodeFromPath(path)
            t = node.data(COLUMN_SIDS)
            d = node.data(COLUMN_DATATYPE)
            v = node.sidsValue()
            if isinstance(v, numpy.ndarray):
                if v.dtype.char in ['S', 'c']:
                    s = '"' + v.tostring().decode('ascii') + '"'
                else:
                    s = str(v.tolist())
            else:
                s = str(v)
            n += '("%s","%s","%s",%s),\n' % (path, t, d, s)
        n += ']\n'
        filename = QFileDialog.getSaveFileName(self,
                                               "Save selection list", ".", "*.py")
        if filename[0] == "":
            return
        with open(str(filename[0]), 'w+') as f:
            f.write(n)

    def deletelocal(self):
        pass

    def reset(self):
        tlvcolsnames = ['Path', 'Name', 'SIDS type', 'DT', 'Value']
        self._tb.clear()
        self._tb.setColumnCount(len(tlvcolsnames))
        self._tb.setRowCount(0)
        lh = self._tb.horizontalHeader()
        lv = self._tb.verticalHeader()
        for i, colname in enumerate(tlvcolsnames):
            hi = QTableWidgetItem(colname)
            self._tb.setHorizontalHeaderItem(i, hi)
        for path in self._data:
            self._tb.setRowCount(self._tb.rowCount() + 1)
            r = self._tb.rowCount() - 1
            self.updateRowData(r, path)
        self.colCheck()
        self._tb.resizeColumnsToContents()
        self._tb.resizeRowsToContents()
        self._parent.treeview.refreshView()

    def updateRowData(self, r, path):
        it0 = QTableWidgetItem('%s ' % '/'.join(path.split('/')[:-1]))
        it0.setFont(QFont("Courier"))
        self._tb.setItem(r, 0, it0)
        it1 = QTableWidgetItem('%s ' % path.split('/')[-1])
        it1.parent = self
        ft = QFont("Courier")
        ft.setWeight(QFont.Bold)
        it1.setFont(ft)
        self._tb.setItem(r, 1, it1)
        node = self._model.nodeFromPath(path)
        if node:
            it1 = QTableWidgetItem(node.data(COLUMN_SIDS))
            it1.parent = self
            it2 = QTableWidgetItem(node.data(COLUMN_DATATYPE))
            it2.parent = self
            val = node.data(COLUMN_VALUE)
            if val == HIDEVALUE:
                val = QIcon(QPixmap(":/images/icons/data-array-large.png"))
                it3 = QTableWidgetItem(val, '')
            else:
                it3 = QTableWidgetItem(val)
                it3.setFont(QFont("Courier"))
            it3.parent = self
            it1.setFont(QFont("Courier"))
            it2.setFont(QFont("Courier"))
            it3.setFont(QFont("Courier"))
            self._tb.setItem(r, 2, it1)
            self._tb.setItem(r, 3, it2)
            self._tb.setItem(r, 4, it3)

    def reject(self):
        self.close()


# -----------------------------------------------------------------
class Q7Query(Q7Window, Ui_Q7QueryWindow):
    _allQueries = {}

    def __init__(self, control, fgindex, treeview):
        Q7Window.__init__(self, Q7Window.VIEW_QUERY, control, '/', fgindex)
        self.bClose.clicked.connect(self.reject)
        self.bRun.clicked.connect(self.runCurrent)
        self.bAdd.clicked.connect(self.queryAdd)
        self.bDel.clicked.connect(self.queryDel)
        self.bRevert.clicked.connect(self.queryRevert)
        self.bCommitDoc.clicked.connect(self.queryCommit)
        self.bRevertDoc.clicked.connect(self.queryRevert)
        self.bSave.clicked.connect(self.queriesSave)
        self.bSaveAsScript.clicked.connect(self.queryScriptSave)
        self._master = treeview
        #
        self.cQueryName.currentIndexChanged.connect(self.changeCurrentQuery)
        self.cQueryGroup.currentIndexChanged.connect(self.changeCurrentGroup)
        #
        # QObject.connect(self.cQueryName,
        #                SIGNAL("currentIndexChanged(int)"),
        #                self.changeCurrentQuery)
        # QObject.connect(self.cQueryGroup,
        #                SIGNAL("currentIndexChanged(int)"),
        #                self.changeCurrentGroup)
        self.cQueryName.editTextChanged.connect(self.checkNewQueryName)
        #
        self.eText.textChanged.connect(self.changeText)
        self.eQueryDoc.textChanged.connect(self.changeText)
        #
        # QObject.connect(self.cQueryName,
        #                SIGNAL("editTextChanged()"),
        #                self.checkNewQueryName)
        # QObject.connect(self.cQueryName,
        #                SIGNAL("editTextChanged(QString)"),
        #                self.checkNewQueryName)
        # QObject.connect(self.eText,
        #                SIGNAL("textChanged()"),
        #                self.changeText)
        # QObject.connect(self.eQueryDoc,
        #                SIGNAL("textChanged()"),
        #                self.changeText)
        self.bInfo.clicked.connect(self.infoQueryView)
        self.bAdd.setEnabled(False)
        self.setCurrentQuery()
        self._modified = False
        self.showQuery()

    def doRelease(self):
        pass

    def infoQueryView(self):
        self._control.helpWindow('Query')

    def updateTreeStatus(self):
        print('query up')

    def queriesSave(self):
        self.queryCommit()
        Q7Query.saveUserQueries()
        self._modified = False

    def queryScriptSave(self):
        n = self.cQueryName.currentText()
        q = self.getQuery(n)
        c = self.eText.toPlainText()
        v = self.eUserVariable.text()
        f = "%s/%s" % (self.FG.filedir, self.FG.filename)
        s = q.getFullScript(f, c, v)
        filename = QFileDialog.getSaveFileName(self,
                                               "Save query script", ".", "*.py")
        if filename[0] == "":
            return
        with open(str(filename[0]), 'w+') as f:
            f.write(s)
        os.chmod(filename[0], 0o0750)
        self._modified = False

    def queryDel(self):
        self._modified = True
        q = self.cQueryName.currentText()
        i = self.cQueryName.currentIndex()
        reply = MSG.message('Delete query',
                            """Do you really want to delete query [%s]?""" % q,
                            MSG.YESNO)
        if reply == QMessageBox.Yes:
            del Q7Query._allQueries[q]
            self.cQueryName.removeItem(i)

    def checkNewQueryName(self, *args):
        qname = self.cQueryName.currentText().strip()
        if qname not in Q7Query._allQueries:
            self.bAdd.setEnabled(True)

    def queryAdd(self):
        self.bAdd.setEnabled(False)
        qname = self.cQueryName.currentText().strip()
        if qname in Q7Query._allQueries:
            return
        q = Q7QueryEntry(qname)
        Q7Query._allQueries[qname] = q
        self.queryCommit()
        self.cQueryName.addItem(qname)

    def queryCommit(self):
        self._modified = True
        q = self.cQueryName.currentText()
        if q not in Q7Query._allQueries:
            self.queryAdd()
        com = self.eText.toPlainText()
        doc = self.eQueryDoc.toPlainText()
        if self.cRequireUpdate.checkState() == Qt.Checked:
            self.setRequireTreeUpdate(True)
        else:
            self.setRequireTreeUpdate(False)
        Q7Query._allQueries[q].setScript(com)
        Q7Query._allQueries[q].setDoc(doc)
        self.bRevert.setEnabled(False)
        self.bCommitDoc.setEnabled(False)
        self.bRevertDoc.setEnabled(False)

    def queryRevert(self):
        self.showQuery()

    def reject(self):
        reply = True
        if self._modified:
            reply = MSG.wQuestion(self._control, 370, 'Leave query panel',
                                  """There are unsaved modified queries,
  leave this panel without save?""", again=False)
        if not reply:
            return
        else:
            self._modified = False
        if self._master._querywindow is not None:
            self._master._querywindow = None
        self.close()

    def reset(self):
        gq = set()
        for qn in self.queriesNamesList():
            self.cQueryName.addItem(qn)
            gq.add(Q7Query._allQueries[qn].group)
        self.cQueryGroup.addItem('*')
        for gqn in gq:
            self.cQueryGroup.addItem(gqn)

    def showQuery(self, name=None):
        pstate = self._modified
        if name is None:
            name = self.getCurrentQuery().name
        if name in Q7Query.queriesNamesList():
            txt = self.getCurrentQuery().script
            self.eText.initText(txt)
            doc = self.getCurrentQuery().doc
            self.eQueryDoc.initText(doc)
        if self.getCurrentQuery().requireTreeUpdate():
            self.cRequireUpdate.setCheckState(Qt.Checked)
        else:
            self.cRequireUpdate.setCheckState(Qt.Unchecked)
        self.bRevert.setEnabled(False)
        self.bCommitDoc.setEnabled(False)
        self.bRevertDoc.setEnabled(False)
        self._modified = pstate

    def show(self):
        self.reset()
        super(Q7Query, self).show()

    def runCurrent(self):
        com = self.eText.toPlainText()
        v = self.eUserVariable.text()
        q = Q7QueryEntry('__tmp__query__')
        q.setScript(com)
        skp = list(self.FG.lazy)
        r = q.run(self.FG.tree, self.FG.links, skp, False, v,
                  self.FG.model.getSelected())
        self.eResult.initText(str(r))
        self.FG.model.markExtendToList(r)
        self.FG.model.updateSelected()
        if q.requireTreeUpdate():
            self.FG.model.modelReset()

        self._master.treeview.refreshView()

    @classmethod
    def fillQueries(cls):
        allqueriestext = Q7Query._userQueriesText + OCTXT._UsualQueries
        Q7Query._defaultQueriesNames = [n[0] for n in OCTXT._UsualQueries]
        for qe in allqueriestext:
            try:
                q = Q7QueryEntry(qe[0], qe[1], qe[2], qe[3],
                                 hasargs=qe[5], update=qe[4])
                Q7Query._allQueries[qe[0]] = q
            except IndexError:
                pass

    @classmethod
    def getQuery(cls, name):
        if name in Q7Query.queriesNamesList():
            return Q7Query._allQueries[name]
        return None

    @classmethod
    def loadUserFunctions(cls):
        try:
            Q7Query._userFunction = OCTXT._readFunctions(cls)()
        except:
            Q7Query._userFunction = None

    @classmethod
    def loadUserQueries(cls):
        Q7Query._userQueriesText = OCTXT._readQueries(cls)
        if Q7Query._userQueriesText is None:
            Q7Query._userQueriesText = []
        return Q7Query._userQueriesText

    @classmethod
    def saveUserQueries(cls):
        ql = []
        for q in Q7Query._allQueries:
            if q not in Q7Query._defaultQueriesNames:
                ql += [str(Q7Query._allQueries[q])]
        OCTXT._writeQueries(cls, ql)

    @classmethod
    def queriesNamesList(cls):
        k = list(Q7Query._allQueries)
        k.sort()
        return k

    @classmethod
    def queriesGroupsList(cls):
        gl = set()
        for q in Q7Query._allQueries:
            g = Q7Query._allQueries[q].group
            gl.add(g)
        return ['*'] + list(gl)

    def getCurrentQuery(self):
        return self._currentQuery

    def changeCurrentGroup(self):
        group = self.cQueryGroup.currentText()
        self.cQueryName.clear()
        for qn in self.queriesNamesList():
            gp = Q7Query._allQueries[qn].group
            if (group == '*') or (gp == group):
                self.cQueryName.addItem(qn)
        self.showQuery()

    def changeCurrentQuery(self, *args):
        name = self.cQueryName.currentText()
        if name not in self.queriesNamesList():
            return
        self.setCurrentQuery(name)
        self.showQuery()

    def setCurrentQuery(self, name=None):
        if name is None:
            name = self.queriesNamesList()[0]
        self._currentQuery = Q7Query._allQueries[name]

    def changeText(self):
        self.bRevert.setEnabled(True)
        self.bCommitDoc.setEnabled(True)
        self.bRevertDoc.setEnabled(True)
        self._modified = True

    @classmethod
    def queries(cls):
        return Q7Query.queriesNamesList()

# -----------------------------------------------------------------



#  ---------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System - 
#  See license.txt file in the root directory of this Python module source  
#  ---------------------------------------------------------------------------
#  
"""
 TYPES, ENUMERATES, CONSTANTS, NAMES from CGNS/MLL

 Conventions:

 [1] A CGNS/SIDS string constant is postfixed with _s
 'ZoneType' is ZoneType_s

 [2] A CGNS/SIDS string constant naming a type has _ts
 'ZoneType_t' is ZoneType_ts

 [3] A list of possible values for a given type has _l
 ZoneType_l is [Null_s,UserDefined_s,Structured_s,Unstructured_s]
 which is same as ["Null","UserDefined","Structured","Unstructured"]
 List should be ordered wrt the actual enumerate

 [4] An enumerate mapping of a list of values is not prefixed
 ZoneType is {'Unstructured':3,'Null':0,'Structured':2,'UserDefined':1}

 [5] The reverse dictionnary of the previous one is postfixed with _
 ZoneType_ is {0:'Null',1:'UserDefined',2:'Structured',3:'Unstructured'}

 [6] The variables are declared with an integer value (not enumerates)
 wrt their position in the _l list, for example:
 (Null,UserDefined,Structured,Unstructured)=list(ZoneType_)

 [7] The _t type names are reserved for Cython, enums are then used as int:
 ctypedef int DataType_t
 int cg_array_read_as(int A, DataType_t type, void *Data)
 
"""
from __future__ import unicode_literals
from builtins import (str, bytes, range, dict)

# ----------------------------------------------------------------------------

def stringAsKeyDict(l):
    return dict(zip(l, range(len(l))))


def enumAsKeyDict(l):
    return dict(zip(range(len(l)), l))


# --------------------------------------------------
# --- ADF-level Datatypes
#
# MT: No data; LK: Link
adftypes = ('C1', 'I4', 'I8', 'R4', 'R8', 'MT', 'LK')
(C1, I4, I8, R4, R8, MT, LK) = adftypes

# --------------------------------------------------
# --- ADF-level Constants
#
ADF_DATA_TYPE_LENGTH = 32
ADF_DATE_LENGTH = 32
ADF_FILENAME_LENGTH = 1024
ADF_FORMAT_LENGTH = 20
ADF_LABEL_LENGTH = 32
ADF_MAXIMUM_LINK_DEPTH = 100
ADF_MAX_DIMENSIONS = 12
ADF_MAX_ERROR_STR_LENGTH = 80
ADF_MAX_LINK_DATA_SIZE = 4096
ADF_NAME_LENGTH = 32
ADF_STATUS_LENGTH = 32
ADF_VERSION_LENGTH = 32

ADF_ROOT_NODE_NAME = "HDF5 MotherNode"
ADF_ROOT_NODE_LABEL = "Root Node of HDF5 File"

CGNSHDF5ROOT_s = ADF_ROOT_NODE_NAME
CGNSLIBRARYVERSION = 4.0

# -------------------------------------------------- (NOT SIDS)
# --- CGNS/Python mapping extensions
#
CGNSTree_ts = 'CGNSTree_t'
CGNSTree_s = 'CGNSTree'

# --- Type with weird (coming from outer space) names
#
Transform_ts = 'Transform_t'
DiffusionModel_ts = 'DiffusionModel_t'
EquationDimension_ts = 'EquationDimension_t'
InwardNormalIndex_ts = 'InwardNormalIndex_t'
IntIndexDimension_ts = 'IntIndexDimension_t'

# --- Add legacy strings for translation tools
#
Transform_ts2 = '"int[IndexDimension]"'
Transform_ts3 = 'int[IndexDimension]'
DiffusionModel_ts2 = '"int[1+...+IndexDimension]"'
EquationDimension_ts2 = '"int"'
InwardNormalIndex_ts2 = '"int[IndexDimension]"'

weirdSIDStypes = {
    Transform_ts2: IntIndexDimension_ts,
    Transform_ts3: IntIndexDimension_ts,
    DiffusionModel_ts2: DiffusionModel_ts,
    EquationDimension_ts2: EquationDimension_ts,
    InwardNormalIndex_ts2: IntIndexDimension_ts,
}

weirdSIDStypes_ = {
    Transform_ts: Transform_ts2,
    DiffusionModel_ts: DiffusionModel_ts2,
    EquationDimension_ts: EquationDimension_ts2,
    InwardNormalIndex_ts: InwardNormalIndex_ts2,
}

# -------------------------------------------------- (SIDS)
# SIDS
#
Null_s = "Null"
UserDefined_s = "UserDefined"

# --------------------------------------------------
Kilogram_s = "Kilogram"
Gram_s = "Gram"
Slug_s = "Slug"
PoundMass_s = "PoundMass"
MassUnits_l = [Null_s, UserDefined_s,
               Kilogram_s, Gram_s, Slug_s, PoundMass_s]

MassUnits = stringAsKeyDict(MassUnits_l)
MassUnits_ = enumAsKeyDict(MassUnits_l)
(MassUnitsNull, MassUnitsUserDefined,
 Kilogram, Gram, Slug, PoundMass) = list(MassUnits_)

# --------------------------------------------------
Meter_s = "Meter"
Centimeter_s = "Centimeter"
Millimeter_s = "Millimeter"
Foot_s = "Foot"
Inch_s = "Inch"
LengthUnits_l = [Null_s, UserDefined_s,
                 Meter_s, Centimeter_s, Millimeter_s, Foot_s, Inch_s]

LengthUnits = stringAsKeyDict(LengthUnits_l)
LengthUnits_ = enumAsKeyDict(LengthUnits_l)
(LengthUnitsNull, LengthUnitsUserDefined,
 Meter, Centimeter, Millimeter, Foot, Inch) = list(LengthUnits_)

# --------------------------------------------------
Second_s = "Second"
TimeUnits_l = [Null_s, UserDefined_s, Second_s]

TimeUnits = stringAsKeyDict(TimeUnits_l)
TimeUnits_ = enumAsKeyDict(TimeUnits_l)
(TimeUnitsNull, TimeUnitsUserDefined, Second) = list(TimeUnits_)

# --------------------------------------------------
Kelvin_s = "Kelvin"
Celcius_s = "Celcius"
Rankine_s = "Rankine"
Fahrenheit_s = "Fahrenheit"
TemperatureUnits_l = [Null_s, UserDefined_s,
                      Kelvin_s, Celcius_s, Rankine_s, Fahrenheit_s]

TemperatureUnits = stringAsKeyDict(TemperatureUnits_l)
TemperatureUnits_ = enumAsKeyDict(TemperatureUnits_l)
(TemperatureUnitsNull, TemperatureUnitsUserDefined,
 Kelvin, Celcius, Rankine, Fahrenheit) = list(TemperatureUnits_)

# --------------------------------------------------
Degree_s = "Degree"
Radian_s = "Radian"
AngleUnits_l = [Null_s, UserDefined_s, Degree_s, Radian_s]

AngleUnits = stringAsKeyDict(AngleUnits_l)
AngleUnits_ = enumAsKeyDict(AngleUnits_l)
(AngleUnitsNull, AngleUnitsUserDefined, Degree, Radian) = list(AngleUnits_)

# --------------------------------------------------
Ampere_s = "Ampere"
Abampere_s = "Abampere"
Statampere_s = "Statampere"
Edison_s = "Edison"
auCurrent_s = "auCurrent"
ElectricCurrentUnits_l = [Null_s, UserDefined_s,
                          Ampere_s, Abampere_s, Statampere_s,
                          Edison_s, auCurrent_s]

ElectricCurrentUnits = stringAsKeyDict(ElectricCurrentUnits_l)
ElectricCurrentUnits_ = enumAsKeyDict(ElectricCurrentUnits_l)
(ElectricCurrentUnitsNull, ElectricCurrentUnitsUserDefined,
 Ampere, Abampere, Statampere,
 Edison, auCurrent) = list(ElectricCurrentUnits_)

# --------------------------------------------------
Mole_s = "Mole"
Entities_s = "Entities"
StandardCubicFoot_s = "StandardCubicFoot"
StandardCubicMeter_s = "StandardCubicMeter"
SubstanceAmountUnits_l = [Null_s, UserDefined_s,
                          Mole_s, Entities_s,
                          StandardCubicFoot_s, StandardCubicMeter_s]

SubstanceAmountUnits = stringAsKeyDict(SubstanceAmountUnits_l)
SubstanceAmountUnits_ = enumAsKeyDict(SubstanceAmountUnits_l)
(SubstanceAmountUnitsNull, SubstanceAmountUnitsUserDefined,
 Mole, Entities,
 StandardCubicFoot, StandardCubicMeter) = list(SubstanceAmountUnits_)

# --------------------------------------------------
Candela_s = "Candela"
Candle_s = "Candle"
Carcel_s = "Carcel"
Hefner_s = "Hefner"
Violle_s = "Violle"
LuminousIntensityUnits_l = [Null_s, UserDefined_s,
                            Candela_s, Candle_s, Carcel_s, Hefner_s, Violle_s]

LuminousIntensityUnits = stringAsKeyDict(LuminousIntensityUnits_l)
LuminousIntensityUnits_ = enumAsKeyDict(LuminousIntensityUnits_l)
(LuminousIntensityUnitsNull, LuminousIntensityUnitsUserDefined,
 Candela, Candle, Carcel, Hefner, Violle) = list(LuminousIntensityUnits_)

# --------------------------------------------------
DimensionalUnits_s = "DimensionalUnits"
AdditionalUnits_s = "AdditionalUnits"
AdditionalExponents_s = "AdditionalExponents"

AllDimensionalUnits_l = TimeUnits_l + MassUnits_l + LengthUnits_l \
                        + TemperatureUnits_l + AngleUnits_l
AllAdditionalUnits_l = LuminousIntensityUnits_l + SubstanceAmountUnits_l \
                       + ElectricCurrentUnits_l
AllUnits_l = AllDimensionalUnits_l + AllAdditionalUnits_l

# --------------------------------------------------
DataClass_ts = "DataClass_t"
DataClass_s = "DataClass"

Dimensional_s = "Dimensional"
NormalizedByDimensional_s = "NormalizedByDimensional"
NormalizedByUnknownDimensional_s = "NormalizedByUnknownDimensional"
NondimensionalParameter_s = "NondimensionalParameter"
DimensionlessConstant_s = "DimensionlessConstant"

DataClass_l = [Null_s, UserDefined_s, Dimensional_s, NormalizedByDimensional_s,
               NormalizedByUnknownDimensional_s, NondimensionalParameter_s,
               DimensionlessConstant_s]
DataClass = stringAsKeyDict(DataClass_l)
DataClass_ = enumAsKeyDict(DataClass_l)
(Null, UserDefined, Dimensional, NormalizedByDimensional,
 NormalizedByUnknownDimensional, NondimensionalParameter,
 DimensionlessConstant) = list(DataClass_)

# ------------------------------------------------------------
GridLocation_ts = "GridLocation_t"
GridLocation_s = "GridLocation"

Vertex_s = "Vertex"
Edge_s = "Edge"
Face_s = "Face"
Cell_s = "Cell"
CellCenter_s = "CellCenter"
FaceCenter_s = "FaceCenter"
IFaceCenter_s = "IFaceCenter"
JFaceCenter_s = "JFaceCenter"
KFaceCenter_s = "KFaceCenter"
EdgeCenter_s = "EdgeCenter"

GridLocation_l = [Null_s, UserDefined_s, Vertex_s, CellCenter_s, FaceCenter_s,
                  IFaceCenter_s, JFaceCenter_s, KFaceCenter_s,
                  EdgeCenter_s]
GridLocation = stringAsKeyDict(GridLocation_l)
GridLocation_ = enumAsKeyDict(GridLocation_l)
(Null, UserDefined, Vertex, CellCenter, FaceCenter,
 IFaceCenter, JFaceCenter, KFaceCenter, EdgeCenter) = list(GridLocation_)

VertexSize_s = "VertexSize"
CellSize_s = "CellSize"
VertexSizeBoundary_s = "VertexSizeBoundary"
ElementsSize_s = "ElementsSize"

# ------------------------------------------------------------
PointSetType_ts = "PointSetType_t"

PointList_s = "PointList"
PointListDonor_s = "PointListDonor"
PointRange_s = "PointRange"
PointRangeDonor_s = "PointRangeDonor"
ElementRange_s = "ElementRange"
ElementList_s = "ElementList"
CellListDonor_s = "CellListDonor"

PointSetType_l = [Null_s, UserDefined_s,
                  PointList_s, PointListDonor_s, PointRange_s, PointRangeDonor_s,
                  ElementRange_s, ElementList_s, CellListDonor_s]
PointSetType = stringAsKeyDict(PointSetType_l)
PointSetType_ = enumAsKeyDict(PointSetType_l)
(Null, UserDefined, PointList, PointListDonor, PointRange, PointRangeDonor,
 ElementRange, ElementList, CellListDonor) = list(PointSetType_)

ZoneDonorName_s = "ZoneDonorName"

# ------------------------------------------------------------
BCDataType_ts = "BCDataType_t"
BCDataType_s = "BCDataType"

DirichletData_s = "DirichletData"
NeumannData_s = "NeumannData"
Dirichlet_s = "Dirichlet"
Neumann_s = "Neumann"

BCDataType_l = [Null_s, UserDefined_s, Dirichlet_s, Neumann_s]
BCDataType = stringAsKeyDict(BCDataType_l)
BCDataType_ = enumAsKeyDict(BCDataType_l)
(BCDataTypeNull, BCDataTypeUserDefined, Dirichlet, Neumann) = list(BCDataType_)

FullPotential_s = "FullPotential"
Euler_s = "Euler"
NSLaminar_s = "NSLaminar"
NSTurbulent_s = "NSTurbulent"
NSLaminarIncompressible_s = "NSLaminarIncompressible"
NSTurbulentIncompressible_s = "NSTurbulentIncompressible"

Ideal_s = "Ideal"
VanderWaals_s = "VanderWaals"
Constant_s = "Constant"
PowerLaw_s = "PowerLaw"
SutherlandLaw_s = "SutherlandLaw"
ConstantPrandtl_s = "ConstantPrandtl"
EddyViscosity_s = "EddyViscosity"
ReynoldsStress_s = "ReynoldsStress"
Algebraic_s = "Algebraic"
BaldwinLomax_s = "BaldwinLomax"
ReynoldsStressAlgebraic_s = "ReynoldsStressAlgebraic"
Algebraic_BaldwinLomax_s = "Algebraic_BaldwinLomax"
Algebraic_CebeciSmith_s = "Algebraic_CebeciSmith"
HalfEquation_JohnsonKing_s = "HalfEquation_JohnsonKing"
OneEquation_BaldwinBarth_s = "OneEquation_BaldwinBarth"
OneEquation_SpalartAllmaras_s = "OneEquation_SpalartAllmaras"
TwoEquation_JonesLaunder_s = "TwoEquation_JonesLaunder"
TwoEquation_MenterSST_s = "TwoEquation_MenterSST"
TwoEquation_Wilcox_s = "TwoEquation_Wilcox"
CaloricallyPerfect_s = "CaloricallyPerfect"
ThermallyPerfect_s = "ThermallyPerfect"
ConstantDensity_s = "ConstantDensity"
RedlichKwong_s = "RedlichKwong"
Frozen_s = "Frozen"
ThermalEquilib_s = "ThermalEquilib"
ThermalNonequilib_s = "ThermalNonequilib"
ChemicalEquilibCurveFit_s = "ChemicalEquilibCurveFit"
ChemicalEquilibMinimization_s = "ChemicalEquilibMinimization"
ChemicalNonequilib_s = "ChemicalNonequilib"
EMElectricField_s = "EMElectricField"
EMMagneticField_s = "EMMagneticField"
EMConductivity_s = "EMConductivity"
Voltage_s = "Voltage"
Interpolated_s = "Interpolated"
Equilibrium_LinRessler_s = "Equilibrium_LinRessler"
Chemistry_LinRessler_s = "Chemistry_LinRessler"

FamilySpecified_s = "FamilySpecified"

# --------------------------------------------------
DataType_ts = "DataType_t"
DataType_s = "DataType"

Integer_s = "Integer"
LongInteger_s = "LongInteger"
RealSingle_s = "RealSingle"
RealDouble_s = "RealDouble"
Character_s = "Character"

DataType_l = [Null_s, UserDefined_s,
              Integer_s, RealSingle_s, RealDouble_s, Character_s, LongInteger_s]
DataType = stringAsKeyDict(DataType_l)
DataType_ = enumAsKeyDict(DataType_l)

(DataTypeNull, DataTypeUserDefined,
 Integer, RealSingle, RealDouble, Character, LongInteger) = list(DataType_)

# --------------------------------------------------
GridConnectivityType_ts = "GridConnectivityType_t"
GridConnectivityType_s = "GridConnectivityType"
GridConnectivity_ts = "GridConnectivity_t"
ZoneGridConnectivity_ts = "ZoneGridConnectivity_t"
ZoneGridConnectivity_s = "ZoneGridConnectivity"

Overset_s = "Overset"
Abutting_s = "Abutting"
Abutting1to1_s = "Abutting1to1"

GridConnectivityType_l = [Null_s, UserDefined_s,
                          Overset_s, Abutting_s, Abutting1to1_s]
GridConnectivityType = stringAsKeyDict(GridConnectivityType_l)
GridConnectivityType_ = enumAsKeyDict(GridConnectivityType_l)

(Null, UserDefined,
 Overset, Abutting, Abutting1to1) = list(GridConnectivityType_)

# --------------------------------------------------
ZoneType_ts = "ZoneType_t"
ZoneType_s = "ZoneType"
Zone_ts = "Zone_t"

Structured_s = "Structured"
Unstructured_s = "Unstructured"

ZoneType_l = [Null_s, UserDefined_s, Structured_s, Unstructured_s]
ZoneType = stringAsKeyDict(ZoneType_l)
ZoneType_ = enumAsKeyDict(ZoneType_l)

(ZoneTypeNull, ZoneTypeUserdefined, Structured, Unstructured) = list(ZoneType_)

ZoneSubRegion_ts = "ZoneSubRegion_t"
BCRegionName_s = "BCRegionName"
GridConnectivityRegionName_s = "GridConnectivityRegionName"

# --------------------------------------------------
SimulationType_ts = "SimulationType_t"
SimulationType_s = "SimulationType"

TimeAccurate_s = "TimeAccurate"
NonTimeAccurate_s = "NonTimeAccurate"

SimulationType_l = [Null_s, UserDefined_s, TimeAccurate_s, NonTimeAccurate_s]
SimulationType = stringAsKeyDict(SimulationType_l)
SimulationType_ = enumAsKeyDict(SimulationType_l)

(Null, UserDefined, TimeAccurate, NonTimeAccurate) = list(SimulationType_)

# --------------------------------------------------
RigidGridMotionType_s = "RigidGridMotionType"
RigidGridMotionType_ts = "RigidGridMotionType_t"

ConstantRate_s = "ConstantRate"
VariableRate_s = "VariableRate"

RigidGridMotionType_l = [Null_s, UserDefined_s, ConstantRate_s, VariableRate_s]
RigidGridMotionType = stringAsKeyDict(RigidGridMotionType_l)
RigidGridMotionType_ = enumAsKeyDict(RigidGridMotionType_l)

(Null, UserDefined, ConstantRate, VariableRate) = list(RigidGridMotionType_)

ArbitraryGridMotionType_s = "ArbitraryGridMotionType"
ArbitraryGridMotionType_ts = "ArbitraryGridMotionType_t"

NonDeformingGrid_s = "NonDeformingGrid"
DeformingGrid_s = "DeformingGrid"

ArbitraryGridMotion_ts = "ArbitraryGridMotion_t"
ArbitraryGridMotion_s = "ArbitraryGridMotion"
ArbitraryGridMotionType_l = [Null_s, UserDefined_s,
                             NonDeformingGrid_s, DeformingGrid_s]
ArbitraryGridMotion = stringAsKeyDict(ArbitraryGridMotionType_l)
ArbitraryGridMotion_ = enumAsKeyDict(ArbitraryGridMotionType_l)

(Null, UserDefined, NonDeformingGrid, DeformingGrid) = list(ArbitraryGridMotion_)

# --------------------------------------------------
WallFunction_ts = "WallFunction_t"
WallFunction_s = "WallFunction"

Generic_s = "Generic"

WallFunctionType_ts = "WallFunctionType_t"
WallFunctionType_s = "WallFunctionType"

WallFunctionType_l = [Null_s, UserDefined_s, Generic_s]
WallFunctionType = stringAsKeyDict(WallFunctionType_l)
WallFunctionType_ = enumAsKeyDict(WallFunctionType_l)

(Null, UserDefined, Generic) = list(WallFunctionType_)

# --------------------------------------------------
Area_ts = "Area_t"
Area_s = "Area"

BleedArea_s = "BleedArea"
CaptureArea_s = "CaptureArea"

AreaType_ts = "AreaType_t"
AreaType_s = "AreaType"
AreaType_l = [Null_s, UserDefined_s, BleedArea_s, CaptureArea_s]
AreaType = stringAsKeyDict(AreaType_l)
AreaType_ = enumAsKeyDict(AreaType_l)
(Null, UserDefined, BleedArea, CaptureArea) = list(AreaType_)

# --------------------------------------------------
ZoneBC_ts = "ZoneBC_t"
ZoneBC_s = "ZoneBC"
ZoneIterativeData_ts = "ZoneIterativeData_t"
ZoneIterativeData_s = "ZoneIterativeData"

UserDefinedData_ts = "UserDefinedData_t"

AverageAll_s = "AverageAll"
AverageCircumferential_s = "AverageCircumferential"
AverageRadial_s = "AverageRadial"
AverageI_s = "AverageI"
AverageJ_s = "AverageJ"
AverageK_s = "AverageK"
CGNSLibraryVersion_s = "CGNSLibraryVersion"
CellDimension_s = "CellDimension"
IndexDimension_s = "IndexDimension"
PhysicalDimension_s = "PhysicalDimension"
GridCoordinates_s = "GridCoordinates"
CoordinateNames_s = "CoordinateNames"
CoordinateX_s = "CoordinateX"
CoordinateY_s = "CoordinateY"
CoordinateZ_s = "CoordinateZ"
CoordinateR_s = "CoordinateR"
CoordinateTheta_s = "CoordinateTheta"
CoordinatePhi_s = "CoordinatePhi"
CoordinateNormal_s = "CoordinateNormal"
CoordinateTangential_s = "CoordinateTangential"
CoordinateXi_s = "CoordinateXi"
CoordinateEta_s = "CoordinateEta"
CoordinateZeta_s = "CoordinateZeta"
CoordinateTransform_s = "CoordinateTransform"
InterpolantsDonor_s = "InterpolantsDonor"
ElementConnectivity_s = "ElementConnectivity"
ElementStartOffset_s = "ElementStartOffset"
ParentData_s = "ParentData"
ParentElements_s = "ParentElements"
ParentElementsPosition_s = "ParentElementsPosition"
ElementSizeBoundary_s = "ElementSizeBoundary"
VectorX_ps = "%sX"
VectorY_ps = "%sY"
VectorZ_ps = "%sZ"
VectorTheta_ps = "%sTheta"
VectorPhi_ps = "%sPhi"
VectorMagnitude_ps = "%sMagnitude"
VectorNormal_ps = "%sNormal"
VectorTangential_ps = "%sTangential"
Potential_s = "Potential"
StreamFunction_s = "StreamFunction"
Density_s = "Density"
Pressure_s = "Pressure"
Temperature_s = "Temperature"
EnergyInternal_s = "EnergyInternal"
Enthalpy_s = "Enthalpy"
Entropy_s = "Entropy"
EntropyApprox_s = "EntropyApprox"
DensityStagnation_s = "DensityStagnation"
PressureStagnation_s = "PressureStagnation"
TemperatureStagnation_s = "TemperatureStagnation"
EnergyStagnation_s = "EnergyStagnation"
EnthalpyStagnation_s = "EnthalpyStagnation"
EnergyStagnationDensity_s = "EnergyStagnationDensity"
VelocityX_s = "VelocityX"
VelocityY_s = "VelocityY"
VelocityZ_s = "VelocityZ"
VelocityR_s = "VelocityR"
VelocityTheta_s = "VelocityTheta"
VelocityPhi_s = "VelocityPhi"
VelocityMagnitude_s = "VelocityMagnitude"
VelocityNormal_s = "VelocityNormal"
VelocityTangential_s = "VelocityTangential"
VelocitySound_s = "VelocitySound"
VelocitySoundStagnation_s = "VelocitySoundStagnation"
MomentumX_s = "MomentumX"
MomentumY_s = "MomentumY"
MomentumZ_s = "MomentumZ"
MomentumMagnitude_s = "MomentumMagnitude"
RotatingVelocityX_s = "RotatingVelocityX"
RotatingVelocityY_s = "RotatingVelocityY"
RotatingVelocityZ_s = "RotatingVelocityZ"
RotatingMomentumX_s = "RotatingMomentumX"
RotatingMomentumY_s = "RotatingMomentumY"
RotatingMomentumZ_s = "RotatingMomentumZ"
RotatingVelocityMagnitude_s = "RotatingVelocityMagnitude"
RotatingPressureStagnation_s = "RotatingPressureStagnation"
RotatingEnergyStagnation_s = "RotatingEnergyStagnation"
RotatingEnergyStagnationDensity_s = "RotatingEnergyStagnationDensity"
RotatingEnthalpyStagnation_s = "RotatingEnthalpyStagnation"
EnergyKinetic_s = "EnergyKinetic"
PressureDynamic_s = "PressureDynamic"
SoundIntensityDB_s = "SoundIntensityDB"
SoundIntensity_s = "SoundIntensity"
VorticityX_s = "VorticityX"
VorticityY_s = "VorticityY"
VorticityZ_s = "VorticityZ"
VorticityMagnitude_s = "VorticityMagnitude"
SkinFrictionX_s = "SkinFrictionX"
SkinFrictionY_s = "SkinFrictionY"
SkinFrictionZ_s = "SkinFrictionZ"
SkinFrictionMagnitude_s = "SkinFrictionMagnitude"
VelocityAngleX_s = "VelocityAngleX"
VelocityAngleY_s = "VelocityAngleY"
VelocityAngleZ_s = "VelocityAngleZ"
VelocityUnitVectorX_s = "VelocityUnitVectorX"
VelocityUnitVectorY_s = "VelocityUnitVectorY"
VelocityUnitVectorZ_s = "VelocityUnitVectorZ"
MassFlow_s = "MassFlow"
ViscosityKinematic_s = "ViscosityKinematic"
ViscosityMolecular_s = "ViscosityMolecular"
ViscosityEddyDynamic_s = "ViscosityEddyDynamic"
ViscosityEddy_s = "ViscosityEddy"
ThermalConductivity_s = "ThermalConductivity"
PowerLawExponent_s = "PowerLawExponent"
SutherlandLawConstant_s = "SutherlandLawConstant"
TemperatureReference_s = "TemperatureReference"
ViscosityMolecularReference_s = "ViscosityMolecularReference"
ThermalConductivityReference_s = "ThermalConductivityReference"
IdealGasConstant_s = "IdealGasConstant"
SpecificHeatPressure_s = "SpecificHeatPressure"
SpecificHeatVolume_s = "SpecificHeatVolume"
ReynoldsStressXX_s = "ReynoldsStressXX"
ReynoldsStressXY_s = "ReynoldsStressXY"
ReynoldsStressXZ_s = "ReynoldsStressXZ"
ReynoldsStressYY_s = "ReynoldsStressYY"
ReynoldsStressYZ_s = "ReynoldsStressYZ"
ReynoldsStressZZ_s = "ReynoldsStressZZ"
LengthReference_s = "LengthReference"
MolecularWeight_s = "MolecularWeight"
MolecularWeight_ps = "MolecularWeight%s"
HeatOfFormation_s = "HeatOfFormation"
HeatOfFormation_ps = "HeatOfFormation%s"
FuelAirRatio_s = "FuelAirRatio"
ReferenceTemperatureHOF_s = "ReferenceTemperatureHOF"
MassFraction_s = "MassFraction"
MassFraction_ps = "MassFraction%s"
LaminarViscosity_s = "LaminarViscosity"
LaminarViscosity_ps = "LaminarViscosity%s"
ThermalConductivity_ps = "ThermalConductivity%s"
EnthalpyEnergyRatio_s = "EnthalpyEnergyRatio"
CompressibilityFactor_s = "CompressibilityFactor"
VibrationalElectronEnergy_s = "VibrationalElectronEnergy"
VibrationalElectronTemperature_s = "VibrationalElectronTemperature"
SpeciesDensity_s = "SpeciesDensity"
SpeciesDensity_ps = "SpeciesDensity%s"
MoleFraction_s = "MoleFraction"
MoleFraction_ps = "MoleFraction%s"
ElectricFieldX_s = "ElectricFieldX"
ElectricFieldY_s = "ElectricFieldY"
ElectricFieldZ_s = "ElectricFieldZ"
MagneticFieldX_s = "MagneticFieldX"
MagneticFieldY_s = "MagneticFieldY"
MagneticFieldZ_s = "MagneticFieldZ"
CurrentDensityX_s = "CurrentDensityX"
CurrentDensityY_s = "CurrentDensityY"
CurrentDensityZ_s = "CurrentDensityZ"
LorentzForceX_s = "LorentzForceX"
LorentzForceY_s = "LorentzForceY"
LorentzForceZ_s = "LorentzForceZ"
ElectricConductivity_s = "ElectricConductivity"
JouleHeating_s = "JouleHeating"
TurbulentDistance_s = "TurbulentDistance"
TurbulentDistanceIndex_s = "TurbulentDistanceIndex"  # not SIDS
TurbulentEnergyKinetic_s = "TurbulentEnergyKinetic"
TurbulentEnergyKineticDensity_s = "TurbulentEnergyKineticDensity"  # not SIDS
TurbulentDissipation_s = "TurbulentDissipation"
TurbulentDissipationRate_s = "TurbulentDissipationRate"
TurbulentDissipationDensity_s = "TurbulentDissipationDensity"  # not SIDS
TurbulentBBReynolds_s = "TurbulentBBReynolds"
TurbulentSANuTilde_s = "TurbulentSANuTilde"
TurbulentSANuTildeDensity_s = "TurbulentSANuTildeDensity"  # not SIDS
Mach_s = "Mach"
Mach_Velocity_s = "Mach_Velocity"
Mach_VelocitySound_s = "Mach_VelocitySound"
Reynolds_s = "Reynolds"
Reynolds_Velocity_s = "Reynolds_Velocity"
Reynolds_Length_s = "Reynolds_Length"
Reynolds_ViscosityKinematic_s = "Reynolds_ViscosityKinematic"
Prandtl_s = "Prandtl"
Prandtl_ThermalConductivity_s = "Prandtl_ThermalConductivity"
Prandtl_ViscosityMolecular_s = "Prandtl_ViscosityMolecular"
Prandtl_SpecificHeatPressure_s = "Prandtl_SpecificHeatPressure"
PrandtlTurbulent_s = "PrandtlTurbulent"
SpecificHeatRatio_s = "SpecificHeatRatio"
SpecificHeatRatio_Pressure_s = "SpecificHeatRatio_Pressure"
SpecificHeatRatio_Volume_s = "SpecificHeatRatio_Volume"
CoefPressure_s = "CoefPressure"
CoefSkinFrictionX_s = "CoefSkinFrictionX"
CoefSkinFrictionY_s = "CoefSkinFrictionY"
CoefSkinFrictionZ_s = "CoefSkinFrictionZ"
Coef_PressureDynamic_s = "Coef_PressureDynamic"
Coef_PressureReference_s = "Coef_PressureReference"
Vorticity_s = "Vorticity"
Acoustic_s = "Acoustic"
RiemannInvariantPlus_s = "RiemannInvariantPlus"
RiemannInvariantMinus_s = "RiemannInvariantMinus"
CharacteristicEntropy_s = "CharacteristicEntropy"
CharacteristicVorticity1_s = "CharacteristicVorticity1"
CharacteristicVorticity2_s = "CharacteristicVorticity2"
CharacteristicAcousticPlus_s = "CharacteristicAcousticPlus"
CharacteristicAcousticMinus_s = "CharacteristicAcousticMinus"
ForceX_s = "ForceX"
ForceY_s = "ForceY"
ForceZ_s = "ForceZ"
ForceR_s = "ForceR"
ForceTheta_s = "ForceTheta"
ForcePhi_s = "ForcePhi"
Lift_s = "Lift"
Drag_s = "Drag"
MomentX_s = "MomentX"
MomentY_s = "MomentY"
MomentZ_s = "MomentZ"
MomentR_s = "MomentR"
MomentTheta_s = "MomentTheta"
MomentPhi_s = "MomentPhi"
MomentXi_s = "MomentXi"
MomentEta_s = "MomentEta"
MomentZeta_s = "MomentZeta"
Moment_CenterX_s = "Moment_CenterX"
Moment_CenterY_s = "Moment_CenterY"
Moment_CenterZ_s = "Moment_CenterZ"
CoefLift_s = "CoefLift"
CoefDrag_s = "CoefDrag"
CoefMomentX_s = "CoefMomentX"
CoefMomentY_s = "CoefMomentY"
CoefMomentZ_s = "CoefMomentZ"
CoefMomentR_s = "CoefMomentR"
CoefMomentTheta_s = "CoefMomentTheta"
CoefMomentPhi_s = "CoefMomentPhi"
CoefMomentXi_s = "CoefMomentXi"
CoefMomentEta_s = "CoefMomentEta"
CoefMomentZeta_s = "CoefMomentZeta"
Coef_PressureDynamic_s = "Coef_PressureDynamic"
Coef_Area_s = "Coef_Area"
Coef_Length_s = "Coef_Length"
TimeValues_s = "TimeValues"
IterationValues_s = "IterationValues"
NumberOfZones_s = "NumberOfZones"
NumberOfFamilies_s = "NumberOfFamilies"
NumberOfSteps_s = "NumberOfSteps"
DataConversion_s = "DataConversion"

ZonePointers_s = "ZonePointers"
FamilyPointers_s = "FamilyPointers"
RigidGridMotionPointers_s = "RigidGridMotionPointers"
ArbitraryGridMotionPointers_s = "ArbitraryGridMotionPointers"
GridCoordinatesPointers_s = "GridCoordinatesPointers"
FlowSolutionPointers_s = "FlowSolutionPointers"
ZoneGridConnectivityPointers_s = "ZoneGridConnectivityPointers"
ZoneSubRegionPointers_s = "ZoneSubRegionPointers"
PointerNames_l = [ZonePointers_s, FamilyPointers_s, RigidGridMotionPointers_s,
                  ArbitraryGridMotionPointers_s, GridCoordinatesPointers_s,
                  FlowSolutionPointers_s,
                  ZoneGridConnectivityPointers_s, ZoneSubRegionPointers_s]

OriginLocation_s = "OriginLocation"
RigidRotationAngle_s = "RigidRotationAngle"
Translation_s = "Translation"
RotationAngle_s = "RotationAngle"
RigidVelocity_s = "RigidVelocity"
RigidRotationRate_s = "RigidRotationRate"
GridVelocityX_s = "GridVelocityX"
GridVelocityY_s = "GridVelocityY"
GridVelocityZ_s = "GridVelocityZ"
GridVelocityR_s = "GridVelocityR"
GridVelocityTheta_s = "GridVelocityTheta"
GridVelocityPhi_s = "GridVelocityPhi"
GridVelocityXi_s = "GridVelocityXi"
GridVelocityEta_s = "GridVelocityEta"
GridVelocityZeta_s = "GridVelocityZeta"

SurfaceArea_s = "SurfaceArea"
RegionName_s = "RegionName"
AverageInterface_ts = "AverageInterface_t"
Axisymmetry_ts = "Axisymmetry_t"
Axisymmetry_s = "Axisymmetry"
AxisymmetryReferencePoint_s = "AxisymmetryReferencePoint"
AxisymmetryAxisVector_s = "AxisymmetryAxisVector"
AxisymmetryAngle_s = "AxisymmetryAngle"
BCDataSet_ts = "BCDataSet_t"
BCData_ts = "BCData_t"
BCData_s = "BCData"

BCProperty_ts = "BCProperty_t"
BCProperty_s = "BCProperty"
BC_ts = "BC_t"

BaseIterativeData_ts = "BaseIterativeData_t"
BaseIterativeData_s = "BaseIterativeData"

CGNSBase_ts = "CGNSBase_t"
CGNSLibraryVersion_ts = "CGNSLibraryVersion_t"

# --------------------------------------------------
ConvergenceHistory_ts = "ConvergenceHistory_t"
ZoneConvergenceHistory_s = "ZoneConvergenceHistory"
GlobalConvergenceHistory_s = "GlobalConvergenceHistory"

ConvergenceHistory_l = [ZoneConvergenceHistory_s,
                        GlobalConvergenceHistory_s]

NormDefinitions_s = "NormDefinitions"

DataArray_ts = "DataArray_t"
DataConversion_ts = "DataConversion_t"
Descriptor_ts = "Descriptor_t"

# --------------------------------------------------
DimensionalExponents_ts = "DimensionalExponents_t"
DimensionalExponents_s = "DimensionalExponents"
DimensionalUnits_ts = "DimensionalUnits_t"
AdditionalUnits_ts = "AdditionalUnits_t"
AdditionalExponents_ts = "AdditionalExponents_t"

DiscreteData_ts = "DiscreteData_t"
DiscreteData_s = "DiscreteData"

FamilyBC_s = "FamilyBC"
FamilyBC_ts = "FamilyBC_t"
FamilyBCDataSet_ts = "FamilyBCDataSet_t"

FamilyName_ts = "FamilyName_t"
FamilyName_s = "FamilyName"
AdditionalFamilyName_ts = "AdditionalFamilyName_t"
AdditionalFamilyName_s = "AdditionalFamilyName"
Family_ts = "Family_t"
Family_s = "Family"
FlowEquationSet_ts = "FlowEquationSet_t"
FlowEquationSet_s = "FlowEquationSet"
FlowSolution_ts = "FlowSolution_t"
GasModel_ts = "GasModel_t"
GasModel_s = "GasModel"

GeometryEntity_ts = "GeometryEntity_t"
GeometryFile_ts = "GeometryFile_t"
GeometryFile_s = "GeometryFile"

GeometryFormat_s = "GeometryFormat"
GeometryFormat_ts = "GeometryFormat_t"

# not supported '-'
NASAIGES_s = "NASA-IGES"
ICEMCFD_s = "ICEM-CFD"

SDRC_s = "SDRC"
Unigraphics_s = "Unigraphics"
ProEngineer_s = "ProEngineer"
GeometryFormat_l = [Null_s, NASAIGES_s, SDRC_s, Unigraphics_s,
                    ProEngineer_s, ICEMCFD_s, UserDefined_s]
GeometryReference_ts = "GeometryReference_t"
GeometryReference_s = "GeometryReference"

Gravity_ts = "Gravity_t"
Gravity_s = "Gravity"
GravityVector_s = "GravityVector"

GridConnectivity1to1_ts = "GridConnectivity1to1_t"
GridConnectivityProperty_ts = "GridConnectivityProperty_t"
GridConnectivityProperty_s = "GridConnectivityProperty"

GridCoordinates_ts = "GridCoordinates_t"
IndexArray_ts = "IndexArray_t"
IndexRange_ts = "IndexRange_t"
IntegralData_ts = "IntegralData_t"
InwardNormalList_ts = "InwardNormalList_t"
InwardNormalList_s = "InwardNormalList"
InwardNormalIndex_s = "InwardNormalIndex"
Ordinal_ts = "Ordinal_t"
Ordinal_s = "Ordinal"
Transform_s = "Transform"
OversetHoles_ts = "OversetHoles_t"
OversetHoles_s = "OversetHoles"
Periodic_ts = "Periodic_t"
Periodic_s = "Periodic"

ReferenceState_ts = "ReferenceState_t"
ReferenceState_s = "ReferenceState"
ReferenceStateDescription_s = "ReferenceStateDescription"

RigidGridMotion_ts = "RigidGridMotion_t"
RigidGridMotion_s = "RigidGridMotion"

Rind_s = "Rind"
Rind_ts = "Rind_t"

RotatingCoordinates_s = "RotatingCoordinates"
RotatingCoordinates_ts = "RotatingCoordinates_t"
RotationRateVector_s = "RotationRateVector"
RotationCenter_s = "RotationCenter"

GoverningEquations_s = "GoverningEquations"
GoverningEquations_ts = "GoverningEquations_t"

GoverningEquationsType_l = [Null_s, UserDefined_s,
                            FullPotential_s, Euler_s, NSLaminar_s,
                            NSTurbulent_s, NSLaminarIncompressible_s,
                            NSTurbulentIncompressible_s]
GoverningEquationsType_s = "GoverningEquationsType"
GoverningEquationsType_ts = "GoverningEquationsType_t"
GoverningEquationsType = stringAsKeyDict(GoverningEquationsType_l)
GoverningEquationsType_ = enumAsKeyDict(GoverningEquationsType_l)
(GoverningEquationsNull, GoverningEquationUserDefined,
 FullPotential, Euler, NSLaminar,
 NSTurbulent, NSLaminarIncompressible,
 NSTurbulentIncompressible) = list(GoverningEquationsType_)

# --------------------------------------------------
BCType_s = "BCType"
BCType_ts = "BCType_t"
BCTypeSimple_s = "BCTypeSimple"
BCTypeSimple_ts = "BCTypeSimple_t"
BCTypeCompound_s = "BCTypeCompound"
BCTypeCompound_ts = "BCTypeCompound_ts"

BCAxisymmetricWedge_s = "BCAxisymmetricWedge"
BCDegenerateLine_s = "BCDegenerateLine"
BCDegeneratePoint_s = "BCDegeneratePoint"
BCDirichlet_s = "BCDirichlet"
BCExtrapolate_s = "BCExtrapolate"
BCFarfield_s = "BCFarfield"
BCGeneral_s = "BCGeneral"
BCInflow_s = "BCInflow"
BCInflowSubsonic_s = "BCInflowSubsonic"
BCInflowSupersonic_s = "BCInflowSupersonic"
BCNeumann_s = "BCNeumann"
BCOutflow_s = "BCOutflow"
BCOutflowSubsonic_s = "BCOutflowSubsonic"
BCOutflowSupersonic_s = "BCOutflowSupersonic"
BCSymmetryPlane_s = "BCSymmetryPlane"
BCSymmetryPolar_s = "BCSymmetryPolar"
BCTunnelInflow_s = "BCTunnelInflow"
BCTunnelOutflow_s = "BCTunnelOutflow"
BCWall_s = "BCWall"
BCWallInviscid_s = "BCWallInviscid"
BCWallViscous_s = "BCWallViscous"
BCWallViscousHeatFlux_s = "BCWallViscousHeatFlux"
BCWallViscousIsothermal_s = "BCWallViscousIsothermal"

BCType_l = [Null_s, UserDefined_s,
            BCAxisymmetricWedge_s, BCDegenerateLine_s, BCDegeneratePoint_s,
            BCDirichlet_s, BCExtrapolate_s, BCFarfield_s,
            BCGeneral_s, BCInflow_s, BCInflowSubsonic_s, BCInflowSupersonic_s,
            BCNeumann_s, BCOutflow_s, BCOutflowSubsonic_s, BCOutflowSupersonic_s,
            BCSymmetryPlane_s, BCSymmetryPolar_s,
            BCTunnelInflow_s, BCTunnelOutflow_s,
            BCWall_s, BCWallInviscid_s, BCWallViscous_s,
            BCWallViscousHeatFlux_s, BCWallViscousIsothermal_s,
            FamilySpecified_s]
BCType = stringAsKeyDict(BCType_l)
BCType_ = enumAsKeyDict(BCType_l)
(Null, UserDefined,
 BCAxisymmetricWedge, BCDegenerateLine, BCDegeneratePoint,
 BCDirichlet, BCExtrapolate, BCFarfield,
 BCGeneral, BCInflow, BCInflowSubsonic, BCInflowSupersonic,
 BCNeumann, BCOutflow, BCOutflowSubsonic, BCOutflowSupersonic,
 BCSymmetryPlane, BCSymmetryPolar,
 BCTunnelInflow, BCTunnelOutflow,
 BCWall, BCWallInviscid, BCWallViscous,
 BCWallViscousHeatFlux, BCWallViscousIsothermal,
 FamilySpecified) = list(BCType_)

BC_l = BCType_l
BC = BCType
BC_ = BCType_

FamilyBC_l = BCType_l
FamilyBC = BCType
FamilyBC_ = BCType_

# CAUTION, index of values in the lists below cannot be used as enumerate,
# the lists are subset of the global list and some index are missing.
BCTypeSimple_l = [Null_s, UserDefined_s, BCGeneral_s, BCDirichlet_s, BCNeumann_s,
                  BCExtrapolate_s, BCWallInviscid_s, BCWallViscousHeatFlux_s,
                  BCWallViscousIsothermal_s, BCWallViscous_s, BCWall_s,
                  BCInflowSubsonic_s, BCInflowSupersonic_s, BCOutflowSubsonic_s,
                  BCOutflowSupersonic_s, BCTunnelInflow_s, BCTunnelOutflow_s,
                  BCDegenerateLine_s, BCDegeneratePoint_s, BCSymmetryPlane_s,
                  BCSymmetryPolar_s, BCAxisymmetricWedge_s, FamilySpecified_s]
BCTypeCompound_l = [Null_s, UserDefined_s,
                    BCInflow_s, BCOutflow_s, BCFarfield_s]

# --------------------------------------------------
# WARNING -
# - the model sub-list CANNOT be used to define the corresponding enumerate
#   because the new values are add at the end of a global list (ModelType)
#   then the ONLY enumerate to use is the ModelType one which has its own list
#
GasModelType_l = [Null_s, UserDefined_s,
                  Ideal_s, VanderWaals_s, CaloricallyPerfect_s,
                  ThermallyPerfect_s, ConstantDensity_s, RedlichKwong_s]
GasModelType_s = "GasModelType"
GasModelType_ts = "GasModelType_t"
GasModelIdentifier_l = [IdealGasConstant_s, SpecificHeatRatio_s,
                        SpecificHeatVolume_s, SpecificHeatPressure_s]

ViscosityModel_ts = "ViscosityModel_t"
ViscosityModel_s = "ViscosityModel"
ViscosityModelType_l = [Null_s, UserDefined_s,
                        Constant_s, PowerLaw_s, SutherlandLaw_s]
ViscosityModelType_s = "ViscosityModelType"
ViscosityModelType_ts = "ViscosityModelType_t"
ViscosityModelIdentifier_l = [PowerLawExponent_s, SutherlandLawConstant_s,
                              TemperatureReference_s,
                              ViscosityMolecularReference_s]

ThermalConductivityModel_ts = "ThermalConductivityModel_t"
ThermalConductivityModel_s = "ThermalConductivityModel"
ThermalConductivityModelType_l = [Null_s, UserDefined_s,
                                  ConstantPrandtl_s, PowerLaw_s,
                                  SutherlandLaw_s]
ThermalConductivityModelType_s = "ThermalConductivityModelType"
ThermalConductivityModelType_ts = "ThermalConductivityModelType_t"
ThermalConductivityModelIdentifier_l = [Prandtl_s, PowerLawExponent_s,
                                        SutherlandLawConstant_s,
                                        TemperatureReference_s,
                                        ThermalConductivityReference_s]

TurbulenceModel_ts = "TurbulenceModel_t"
TurbulenceModel_s = "TurbulenceModel"
TurbulenceModelType_l = [Null_s, UserDefined_s,
                         Algebraic_BaldwinLomax_s,
                         Algebraic_CebeciSmith_s,
                         HalfEquation_JohnsonKing_s,
                         OneEquation_BaldwinBarth_s,
                         OneEquation_SpalartAllmaras_s,
                         TwoEquation_JonesLaunder_s,
                         TwoEquation_MenterSST_s, TwoEquation_Wilcox_s]
TurbulenceModelType_s = "TurbulenceModelType"
TurbulenceModelType_ts = "TurbulenceModelType_t"

TurbulenceClosure_ts = "TurbulenceClosure_t"
TurbulenceClosure_s = "TurbulenceClosure"
TurbulenceClosureType_l = [Null_s, UserDefined_s,
                           EddyViscosity_s, ReynoldsStress_s,
                           ReynoldsStressAlgebraic_s]
TurbulenceClosureType_s = "TurbulenceClosureType"
TurbulenceClosureType_ts = "TurbulenceClosureType_t"
TurbulenceClosureIdentifier_l = [PrandtlTurbulent_s]

ThermalRelaxationModel_ts = "ThermalRelaxationModel_t"
ThermalRelaxationModel_s = "ThermalRelaxationModel"
ThermalRelaxationModelType_l = [Null_s, UserDefined_s,
                                Frozen_s, ThermalEquilib_s,
                                ThermalNonequilib_s]
ThermalRelaxationModelType_s = "ThermalRelaxationModelType"
ThermalRelaxationModelType_ts = "ThermalRelaxationModelType_t"

ChemicalKineticsModel_ts = "ChemicalKineticsModel_t"
ChemicalKineticsModel_s = "ChemicalKineticsModel"
ChemicalKineticsModelType_l = [Null_s, UserDefined_s,
                               Frozen_s, ChemicalEquilibCurveFit_s,
                               ChemicalEquilibMinimization_s,
                               ChemicalNonequilib_s]
ChemicalKineticsModelType_s = "ChemicalKineticsModelType"
ChemicalKineticsModelType_ts = "ChemicalKineticsModelType_t"
ChemicalKineticsModelIdentifier_l = [FuelAirRatio_s, ReferenceTemperatureHOF_s]

EMElectricFieldModel_s = "EMElectricFieldModel"
EMElectricFieldModel_ts = "EMElectricFieldModel_t"
EMElectricFieldModelType_l = [Null_s, UserDefined_s,
                              Constant_s, Frozen_s,
                              Interpolated_s, Voltage_s]
EMElectricFieldModelType_s = "EMElectricFieldModelType"
EMElectricFieldModelType_ts = "EMElectricFieldModelType_t"

EMMagneticFieldModel_s = "EMMagneticFieldModel"
EMMagneticFieldModel_ts = "EMMagneticFieldModel_t"
EMMagneticFieldModelType_l = [Null_s, UserDefined_s,
                              Constant_s, Frozen_s, Interpolated_s]
EMMagneticFieldModelType_s = "EMMagneticFieldModelType"
EMMagneticFieldModelType_ts = "EMMagneticFieldModelType_t"

EMConductivityModel_s = "EMConductivityModel"
EMConductivityModel_ts = "EMConductivityModel_t"
EMConductivityModelType_l = [Null_s, UserDefined_s,
                             Constant_s, Frozen_s,
                             Equilibrium_LinRessler_s,
                             Chemistry_LinRessler_s]
EMConductivityModelType_s = "EMConductivityModelType"
EMConductivityModelType_ts = "EMConductivityModelType_t"
EMConductivityModelIdentifier_l = [Null_s, UserDefined_s,
                                   ElectricFieldX_s, ElectricFieldY_s,
                                   ElectricFieldZ_s, MagneticFieldX_s,
                                   MagneticFieldY_s, MagneticFieldZ_s,
                                   CurrentDensityX_s, CurrentDensityY_s,
                                   CurrentDensityZ_s, ElectricConductivity_s,
                                   LorentzForceX_s, LorentzForceY_s,
                                   LorentzForceZ_s, JouleHeating_s]

ModelType_l = [Null_s, UserDefined_s,
               Ideal_s, VanderWaals_s, Constant_s, PowerLaw_s, SutherlandLaw_s,
               ConstantPrandtl_s, EddyViscosity_s, ReynoldsStress_s,
               ReynoldsStressAlgebraic_s, Algebraic_BaldwinLomax_s,
               Algebraic_CebeciSmith_s, HalfEquation_JohnsonKing_s,
               OneEquation_BaldwinBarth_s, OneEquation_SpalartAllmaras_s,
               TwoEquation_JonesLaunder_s, TwoEquation_MenterSST_s,
               TwoEquation_Wilcox_s, CaloricallyPerfect_s, ThermallyPerfect_s,
               ConstantDensity_s, RedlichKwong_s, Frozen_s, ThermalEquilib_s,
               ThermalNonequilib_s, ChemicalEquilibCurveFit_s,
               ChemicalEquilibMinimization_s, ChemicalNonequilib_s,
               EMElectricField_s, EMMagneticField_s, EMConductivity_s,
               Voltage_s, Interpolated_s, Equilibrium_LinRessler_s,
               Chemistry_LinRessler_s]
ModelType = stringAsKeyDict(ModelType_l)
ModelType_ = enumAsKeyDict(ModelType_l)
(Null, UserDefined,
 Ideal, VanderWaals, Constant, PowerLaw, SutherlandLaw,
 ConstantPrandtl, EddyViscosity, ReynoldsStress,
 ReynoldsStressAlgebraic, Algebraic_BaldwinLomax,
 Algebraic_CebeciSmith, HalfEquation_JohnsonKing,
 OneEquation_BaldwinBarth, OneEquationpalartAllmaras,
 TwoEquation_JonesLaunder, TwoEquation_MenterSST,
 TwoEquation_Wilcox, CaloricallyPerfect, ThermallyPerfect,
 ConstantDensity, RedlichKwong, Frozen, ThermalEquilib,
 ThermalNonequilib, ChemicalEquilibCurveFit,
 ChemicalEquilibMinimization, ChemicalNonequilib,
 EMElectricField, EMMagneticField, EMConductivity,
 Voltage, Interpolated, Equilibrium_LinRessler,
 Chemistry_LinRessler) = list(ModelType_)

DiffusionModel_s = 'DiffusionModel'
EquationDimension_s = 'EquationDimension'

# --------------------------------------------------

AverageInterfaceType_s = "AverageInterfaceType"
AverageInterfaceType_ts = "AverageInterfaceType_t"
AverageInterfaceType_l = [Null_s, UserDefined_s,
                          AverageAll_s, AverageCircumferential_s,
                          AverageRadial_s, AverageI_s, AverageJ_s, AverageK_s]

AverageInterfaceType = stringAsKeyDict(AverageInterfaceType_l)
AverageInterfaceType_ = enumAsKeyDict(AverageInterfaceType_l)
(Null, UserDefined,
 AverageAll, AverageCircumferential,
 AverageRadial, AverageI, AverageJ, AverageK) = list(AverageInterfaceType_)

AverageInterface_s = "AverageInterface"
AverageInterface_ts = "AverageInterface_t"

NODE_s = "NODE"
BAR_2_s = "BAR_2"
BAR_3_s = "BAR_3"
TRI_3_s = "TRI_3"
TRI_6_s = "TRI_6"
QUAD_4_s = "QUAD_4"
QUAD_8_s = "QUAD_8"
QUAD_9_s = "QUAD_9"
TETRA_4_s = "TETRA_4"
TETRA_10_s = "TETRA_10"
PYRA_5_s = "PYRA_5"
PYRA_14_s = "PYRA_14"
PENTA_6_s = "PENTA_6"
PENTA_15_s = "PENTA_15"
PENTA_18_s = "PENTA_18"
HEXA_8_s = "HEXA_8"
HEXA_20_s = "HEXA_20"
HEXA_27_s = "HEXA_27"
MIXED_s = "MIXED"
PYRA_13_s = "PYRA_13"
NGON_n_s = "NGON_n"
NFACE_n_s = "NFACE_n"
BAR_4_s = "BAR_4"
TRI_9_s = "TRI_9"
TRI_10_s = "TRI_10"
QUAD_12_s = "QUAD_12"
QUAD_16_s = "QUAD_16"
TETRA_16_s = "TETRA_16"
TETRA_20_s = "TETRA_20"
PYRA_21_s = "PYRA_21"
PYRA_29_s = "PYRA_29"
PYRA_30_s = "PYRA_30"
PENTA_24_s = "PENTA_24"
PENTA_38_s = "PENTA_38"
PENTA_40_s = "PENTA_40"
HEXA_32_s = "HEXA_32"
HEXA_56_s = "HEXA_56"
HEXA_64_s = "HEXA_64"
BAR_5_s = "BAR_5"
TRI_12_s = "TRI_12"
TRI_15_s = "TRI_15"
QUAD_P4_16_s = "QUAD_P4_16"
QUAD_25_s = "QUAD_25"
TETRA_22_s = "TETRA_22"
TETRA_34_s = "TETRA_34"
TETRA_35_s = "TETRA_35"
PYRA_P4_29_s = "PYRA_P4_29"
PYRA_50_s = "PYRA_50"
PYRA_55_s = "PYRA_55"
PENTA_33_s = "PENTA_33"
PENTA_66_s = "PENTA_66"
PENTA_75_s = "PENTA_75"
HEXA_44_s = "HEXA_44"
HEXA_98_s = "HEXA_98"
HEXA_125_s = "HEXA_125"

# ORDER *IS* SIGNIFICANT

Null_npe = 0
UserDefined_npe = 0
NODE_npe = 1
BAR_2_npe = 2
BAR_3_npe = 3
TRI_3_npe = 3
TRI_6_npe = 6
QUAD_4_npe = 4
QUAD_8_npe = 8
QUAD_9_npe = 9
TETRA_4_npe = 4
TETRA_10_npe = 10
PYRA_5_npe = 5
PYRA_14_npe = 14
PENTA_6_npe = 6
PENTA_15_npe = 15
PENTA_18_npe = 18
HEXA_8_npe = 8
HEXA_20_npe = 20
HEXA_27_npe = 27
MIXED_npe = 0
PYRA_13_npe = 13
NGON_n_npe = 0
NFACE_n_npe = 0
BAR_4_npe = 4
TRI_9_npe = 9
TRI_10_npe = 10
QUAD_12_npe = 12
QUAD_16_npe = 16
TETRA_16_npe = 16
TETRA_20_npe = 20
PYRA_21_npe = 21
PYRA_29_npe = 29
PYRA_30_npe = 30
PENTA_24_npe = 24
PENTA_38_npe = 38
PENTA_40_npe = 40
HEXA_32_npe = 32
HEXA_56_npe = 56
HEXA_64_npe = 64
BAR_5_npe = 40
TRI_12_npe = 41
TRI_15_npe = 42
QUAD_P4_16_npe = 43
QUAD_25_npe = 44
TETRA_22_npe = 45
TETRA_34_npe = 46
TETRA_35_npe = 47
PYRA_P4_29_npe = 48
PYRA_50_npe = 49
PYRA_55_npe = 50
PENTA_33_npe = 51
PENTA_66_npe = 52
PENTA_75_npe = 53
HEXA_44_npe = 54
HEXA_98_npe = 55
HEXA_125_npe = 56

Elements_ts = "Elements_t"
ElementType_ts = "ElementType_t"
ElementType_s = "ElementType"
Elements_s = "Elements"
ElementType_l = [Null_s, UserDefined_s, NODE_s, BAR_2_s, BAR_3_s,
                 TRI_3_s, TRI_6_s, QUAD_4_s, QUAD_8_s, QUAD_9_s,
                 TETRA_4_s, TETRA_10_s, PYRA_5_s, PYRA_14_s,
                 PENTA_6_s, PENTA_15_s, PENTA_18_s,
                 HEXA_8_s, HEXA_20_s, HEXA_27_s, MIXED_s, PYRA_13_s,
                 NGON_n_s, NFACE_n_s, BAR_4_s, TRI_9_s, TRI_10_s, QUAD_12_s,
                 QUAD_16_s, TETRA_16_s, TETRA_20_s, PYRA_21_s, PYRA_29_s,
                 PYRA_30_s, PENTA_24_s, PENTA_38_s, PENTA_40_s, HEXA_32_s,
                 HEXA_56_s, HEXA_64_s, BAR_5_s, TRI_12_s, TRI_15_s,
                 QUAD_P4_16_s, QUAD_25_s, TETRA_22_s, TETRA_34_s,
                 TETRA_35_s, PYRA_P4_29_s, PYRA_50_s, PYRA_55_s, PENTA_33_s,
                 PENTA_66_s, PENTA_75_s, HEXA_44_s, HEXA_98_s, HEXA_125_s]
ElementTypeNPE_l = [Null_npe, UserDefined_npe, NODE_npe, BAR_2_npe, BAR_3_npe,
                    TRI_3_npe, TRI_6_npe, QUAD_4_npe, QUAD_8_npe, QUAD_9_npe,
                    TETRA_4_npe, TETRA_10_npe, PYRA_5_npe, PYRA_14_npe,
                    PENTA_6_npe, PENTA_15_npe, PENTA_18_npe,
                    HEXA_8_npe, HEXA_20_npe, HEXA_27_npe, MIXED_npe,
                    PYRA_13_npe, NGON_n_npe, NFACE_n_npe, BAR_4_npe, TRI_9_npe,
                    TRI_10_npe, QUAD_12_npe, QUAD_16_npe, TETRA_16_npe,
                    TETRA_20_npe, PYRA_21_npe, PYRA_29_npe,
                    PYRA_30_npe, PENTA_24_npe, PENTA_38_npe, PENTA_40_npe,
                    HEXA_32_npe, HEXA_56_npe, HEXA_64_npe,
                    BAR_5_npe, TRI_12_npe, TRI_15_npe,
                    QUAD_P4_16_npe, QUAD_25_npe, TETRA_22_npe, TETRA_34_npe,
                    TETRA_35_npe, PYRA_P4_29_npe, PYRA_50_npe, PYRA_55_npe,
                    PENTA_33_npe,
                    PENTA_66_npe, PENTA_75_npe,
                    HEXA_44_npe, HEXA_98_npe, HEXA_125_npe]
ElementType = stringAsKeyDict(ElementType_l)
ElementType_ = enumAsKeyDict(ElementType_l)
ElementTypeNPE = dict(zip(ElementType_l, ElementTypeNPE_l))
(Null, UserDefined, NODE, BAR_2, BAR_3,
 TRI_3, TRI_6, QUAD_4, QUAD_8, QUAD_9,
 TETRA_4, TETRA_10, PYRA_5, PYRA_14,
 PENTA_6, PENTA_15, PENTA_18,
 HEXA_8, HEXA_20, HEXA_27, MIXED, PYRA_13,
 NGON_n, NFACE_n,
 BAR_4, TRI_9, TRI_10, QUAD_12,
 QUAD_16, TETRA_16, TETRA_20, PYRA_21, PYRA_29,
 PYRA_30, PENTA_24, PENTA_38, PENTA_40, HEXA_32,
 HEXA_56, HEXA_64, BAR_5, TRI_12, TRI_15,
 QUAD_P4_16, QUAD_25, TETRA_22, TETRA_34,
 TETRA_35, PYRA_P4_29, PYRA_50, PYRA_55,
 PENTA_33,
 PENTA_66, PENTA_75,
 HEXA_44, HEXA_98, HEXA_125) = list(ElementType_)

ElementType3D = [TETRA_4, TETRA_10, PYRA_5, PYRA_14,
                 PENTA_6, PENTA_15, PENTA_18,
                 HEXA_8, HEXA_20, HEXA_27, MIXED, PYRA_13,
                 TETRA_16, TETRA_20, PYRA_21, PYRA_29,
                 PYRA_30, PENTA_24, PENTA_38, PENTA_40, HEXA_32,
                 HEXA_56, HEXA_64, TETRA_22, TETRA_34,
                 TETRA_35, PYRA_P4_29, PYRA_50, PYRA_55,
                 PENTA_33, PENTA_66, PENTA_75,
                 HEXA_44, HEXA_98, HEXA_125]
ElementType2D = [TRI_3, TRI_6, QUAD_4, QUAD_8, QUAD_9,
                 TRI_9, TRI_10, QUAD_12, QUAD_16, TRI_12, TRI_15,
                 QUAD_P4_16, QUAD_25]
ElementType1D = [BAR_2, BAR_3, BAR_4, BAR_5]
ElementType0D = [NODE]
ElementType_tri = [TRI_3, TRI_6, TRI_9, TRI_10, TRI_12, TRI_15]
ElementType_quad = [QUAD_4, QUAD_8, QUAD_9, QUAD_12, QUAD_16,
                    QUAD_P4_16, QUAD_25]
ElementType_tetra = [TETRA_4, TETRA_10, TETRA_16, TETRA_20, TETRA_22, TETRA_34,
                     TETRA_35]
ElementType_pyra = [PYRA_5, PYRA_13, PYRA_21, PYRA_29, PYRA_30,
                    PYRA_P4_29, PYRA_50, PYRA_55]
ElementType_penta = [PENTA_6, PENTA_15, PENTA_18, PENTA_24, PENTA_38, PENTA_40,
                     PENTA_33, PENTA_66, PENTA_75]
ElementType_hexa = [HEXA_8, HEXA_20, HEXA_27, HEXA_32, HEXA_56, HEXA_64,
                    HEXA_44, HEXA_98, HEXA_125]
ElementType_trionly = ElementType_tetra
ElementType_quadonly = ElementType_hexa
ElementType_triquad = ElementType_pyra + ElementType_penta

ElementRangeList_s = "ElementRangeList"
#

# ---
cgnsnames = [globals()[k] for k in dir() if (k[-2:] == '_s')]
cgnstypes = [globals()[k] for k in dir() if (k[-3:] == '_ts')] + list(weirdSIDStypes)
cgnsenums = {
    GoverningEquations_ts: GoverningEquationsType_l,
    GasModel_ts: GasModelType_l,
    GasModelType_ts: GasModelIdentifier_l,
    TurbulenceModel_ts: TurbulenceModelType_l,
    ViscosityModel_ts: ViscosityModelType_l,
    ViscosityModelType_ts: ViscosityModelIdentifier_l,
    ThermalConductivityModel_ts: ThermalConductivityModelType_l,
    ThermalConductivityModelType_ts: ThermalConductivityModelIdentifier_l,
    ThermalRelaxationModel_ts: ThermalRelaxationModelType_l,
    ChemicalKineticsModel_ts: ChemicalKineticsModelType_l,
    ChemicalKineticsModelType_ts: ChemicalKineticsModelIdentifier_l,
    EMElectricFieldModel_ts: EMElectricFieldModelType_l,
    EMMagneticFieldModel_ts: EMMagneticFieldModelType_l,
    EMConductivityModel_ts: EMConductivityModelType_l,
    EMConductivityModelType_ts: EMConductivityModelIdentifier_l,
}
for k in dir():
    if (k[-2:] == '_l'):
        cgnsenums[k[:-1] + 't'] = locals()[k]
#
cgnsnames.sort()
cgnstypes.sort()
#
# Compatibility
# - some constants have no integer/string correspondance, the legacy wrapper
#   had its own defines, try to set them as long as it doens't collide with
#   existing constants.
#
CoordinateX = CoordinateX_s
CoordinateY = CoordinateY_s
CoordinateZ = CoordinateZ_s
CoefLift = CoefLift_s
Density = Density_s

Zone_t = Zone_ts
ReferenceState_t = ReferenceState_ts
FlowEquationSet_t = FlowEquationSet_ts
GoverningEquations_t = GoverningEquations_ts
GasModel_t = GasModel_ts
ViscosityModel_t = ViscosityModel_ts
TurbulenceModel_t = TurbulenceModel_ts
UserDefinedData_t = UserDefinedData_ts
FlowSolution_t = FlowSolution_ts
DataArray_t = DataArray_ts
Family_t = Family_ts
DiscreteData_t = DiscreteData_ts
RigidGridMotion_t = RigidGridMotion_ts

# sorted type list, for CGNS.NAV use or any CGNS/SIDS doc related order
sortedtypelist = [
    CGNSTree_ts,
    Family_ts,
    FamilyName_ts,
    AdditionalFamilyName_ts,
    CGNSBase_ts,
    Zone_ts,
    ZoneType_ts,
    GridCoordinates_ts,
    Elements_ts,
    ZoneBC_ts,
    AdditionalExponents_ts,
    AdditionalUnits_ts,
    ArbitraryGridMotionType_ts,
    ArbitraryGridMotion_ts,
    AreaType_ts,
    Area_ts,
    AverageInterfaceType_ts,
    AverageInterface_ts,
    Axisymmetry_ts,
    BCDataSet_ts,
    BCData_ts,
    BCProperty_ts,
    BCTypeSimple_ts,
    BCType_ts,
    BC_ts,
    BaseIterativeData_ts,
    ChemicalKineticsModelType_ts,
    ChemicalKineticsModel_ts,
    ConvergenceHistory_ts,
    DataArray_ts,
    DataClass_ts,
    DataConversion_ts,
    DataType_ts,
    Descriptor_ts,
    DiffusionModel_ts,
    DimensionalExponents_ts,
    DimensionalUnits_ts,
    DiscreteData_ts,
    EMConductivityModelType_ts,
    EMConductivityModel_ts,
    EMElectricFieldModelType_ts,
    EMElectricFieldModel_ts,
    EMMagneticFieldModelType_ts,
    EMMagneticFieldModel_ts,
    ElementType_ts,
    EquationDimension_ts,
    FamilyBC_ts,
    FlowEquationSet_ts,
    FlowSolution_ts,
    GasModelType_ts,
    GasModel_ts,
    GeometryEntity_ts,
    GeometryFile_ts,
    GeometryFormat_ts,
    GeometryReference_ts,
    GoverningEquationsType_ts,
    GoverningEquations_ts,
    Gravity_ts,
    GridConnectivity1to1_ts,
    GridConnectivityProperty_ts,
    GridConnectivityType_ts,
    GridConnectivity_ts,
    GridLocation_ts,
    IndexArray_ts,
    IndexRange_ts,
    IntegralData_ts,
    InwardNormalIndex_ts,
    InwardNormalList_ts,
    Ordinal_ts,
    OversetHoles_ts,
    Periodic_ts,
    ReferenceState_ts,
    RigidGridMotionType_ts,
    RigidGridMotion_ts,
    Rind_ts,
    RotatingCoordinates_ts,
    SimulationType_ts,
    ThermalConductivityModelType_ts,
    ThermalConductivityModel_ts,
    ThermalRelaxationModelType_ts,
    ThermalRelaxationModel_ts,
    Transform_ts,
    TurbulenceClosureType_ts,
    TurbulenceClosure_ts,
    TurbulenceModelType_ts,
    TurbulenceModel_ts,
    UserDefinedData_ts,
    ViscosityModelType_ts,
    ViscosityModel_ts,
    WallFunctionType_ts,
    WallFunction_ts,
    ZoneGridConnectivity_ts,
    ZoneIterativeData_ts,
    CGNSLibraryVersion_ts,
]


#
# --- last line



#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System - 
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
from __future__ import unicode_literals
from builtins import (str, bytes, range, dict)


import os

from qtpy.QtCore import Qt
from qtpy.QtGui import QFont
from qtpy.QtWidgets import QTreeView, QFileDialog, QTreeWidgetItem, QWidget


import CGNS.NAV.wmessages as MSG
import CGNS.VAL.parse.messages as CGM
from CGNS.NAV.Q7DiagWindow import Ui_Q7DiagWindow
from CGNS.NAV.moption import Q7OptionContext as OCTXT
from CGNS.NAV.wfingerprint import Q7Window as QW
from CGNS.VAL.parse.findgrammar import locateGrammars


# -----------------------------------------------------------------
class Q7CheckList(QW, Ui_Q7DiagWindow):
    def __init__(self, parent, data, fgindex):
        QW.__init__(self, QW.VIEW_DIAG, parent._control, None, fgindex)
        self._currentItem = 0
        self._filterItems = {}
        self.bClose.clicked.connect(self.reject)
        self.bExpandAll.clicked.connect(self.expand)
        self.bInfo.clicked.connect(self.infoDiagView)
        self.bCollapseAll.clicked.connect(self.collapse)
        self.bPrevious.clicked.connect(self.previousfiltered)
        self.bNext.clicked.connect(self.nextfiltered)
        self.cWarnings.clicked.connect(self.reset)
        self.cDiagFirst.clicked.connect(self.reset)
        self.bSave.clicked.connect(self.diagnosticssave)
        self.bWhich.clicked.connect(self.grammars)
        self.cFilter.currentIndexChanged[int].connect(self.filterChange)
        # QObject.connect(self.cFilter,
        #                 SIGNAL("currentIndexChanged(int)"),
        #                 self.filterChange)
        self._data = data
        self._parent = parent
        self.diagTable.keyPressEvent = self.diagTableKeyPressEvent
        self.filterChange()

    def doRelease(self):
        pass

    def grammars(self):
        self.FG.pushGrammarPaths()
        gl = ""
        for g in locateGrammars():
            gl += "%s : %s<br>" % (g[0], os.path.split(g[1])[0])
        self.FG.popGrammarPaths()
        MSG.wInfo(self, 400, "Diag view:",
                  """Checks performed using the following grammars:<br>%s""" % gl,
                  again=False)

    def diagTableKeyPressEvent(self, event):
        kmod = event.modifiers()
        kval = event.key()
        if kval == Qt.Key_Space:
            itlist = self.diagTable.selectedItems()
            it = itlist[0]
            itxt = it.text(0)
            if itxt[0] != '/':
                itxt = it.parent().text(0)
            self._parent.treeview.selectByPath(itxt)
        else:
            QTreeView.keyPressEvent(self.diagTable, event)

    def infoDiagView(self):
        self._control.helpWindow('Diagnosis')

    def diagnosticssave(self):
        n = 'data=%s\n' % self._data

        filename = QFileDialog.getSaveFileName(self,
                                               "Save diagnosis", ".", "*.py")
        if filename[0] == "":
            return
        with open(str(filename[0]), 'w+') as f:
            f.write(n)

    def previousfiltered(self):
        iold = self._filterItems[self.cFilter.currentText()][self._currentItem]
        if self._currentItem != 0:
            self._currentItem -= 1
        inew = self._filterItems[self.cFilter.currentText()][self._currentItem]
        iold.setSelected(False)
        inew.setSelected(True)
        self.diagTable.scrollToItem(inew)
        self.eCount.setText(str(self._currentItem + 1))

    def nextfiltered(self):
        ilist = self._filterItems[self.cFilter.currentText()]
        if self._currentItem >= len(ilist) - 1:
            return
        iold = ilist[self._currentItem]
        self._currentItem += 1
        inew = ilist[self._currentItem]
        iold.setSelected(False)
        inew.setSelected(True)
        self.diagTable.scrollToItem(inew)
        self.eCount.setText(str(self._currentItem + 1))

    def filterChange(self):
        self._currentItem = 0
        self.eCount.setText("")
        self.diagTable.clearSelection()

    def expand(self):
        self.diagTable.expandAll()

    def collapse(self):
        self.diagTable.collapseAll()

    def show(self):
        self.reset()
        super(Q7CheckList, self).show()

    def reset(self):
        v = self.diagTable
        v.clear()
        v.setHeaderHidden(True)
        plist = list(self._data)
        plist.sort()
        plist.reverse()
        keyset = set()
        self.cFilter.clear()
        self._filterItems = {}
        diagfirst = self.cDiagFirst.isChecked()
        warnings = self.cWarnings.isChecked()        
        diagstack = {}
        for path in plist:
            path_item = None
            state = self._data.getWorstDiag(path)
            if (state in [CGM.CHECK_NONE, CGM.CHECK_PASS]):
                pass
            elif (state == CGM.CHECK_WARN) and not warnings:
                pass
            else:
                for (diag, pth) in self._data.diagnosticsByPath(path):
                    if (diag.level == CGM.CHECK_WARN) and not warnings:
                        pass
                    elif diagfirst:
                        if diag.key not in keyset:
                            keyset.add(diag.key)
                            diag_item = self.addDiagEntry(None, diag, top=True)
                            v.insertTopLevelItem(0, diag_item)
                            diagstack[diag.key] = diag_item
                        path_item = self.addPathEntry(diagstack[diag.key],
                                                      path,
                                                      state, top=False)
                    else:
                        if diag.key not in keyset:
                            keyset.add(diag.key)
                        if path_item is None:
                            path_item = self.addPathEntry(None, path,
                                                          state, top=True)
                            v.insertTopLevelItem(0, path_item)
                        diag_item = self.addDiagEntry(path_item, diag,
                                                      top=False)
                    keylist = list(keyset)
        keylist.sort()
        for k in keylist:
            self.cFilter.addItem(k)

    def addPathEntry(self, parent, path, state, top=False):
        it = QTreeWidgetItem(parent, (path,))
        ft = QFont(OCTXT._Table_Font)
        if top:
            ft.setBold(True)
        it.setFont(0, ft)
        if state == CGM.CHECK_FAIL:
            it.setIcon(0, self.IC(QW.I_C_SFL))
        if state == CGM.CHECK_WARN:
            it.setIcon(0, self.IC(QW.I_C_SWR))
        return it

    def addDiagEntry(self, parent, diag, top=False):
        dit = QTreeWidgetItem(parent, (self._data.message(diag),))
        ft = QFont(OCTXT._Table_Font)
        if top:
            ft.setBold(True)
        dit.setFont(0, ft)
        if diag.level == CGM.CHECK_FAIL:
            dit.setIcon(0, self.IC(QW.I_C_SFL))
        if diag.level == CGM.CHECK_WARN:
            dit.setIcon(0, self.IC(QW.I_C_SWR))
        if diag.key not in self._filterItems:
            self._filterItems[diag.key] = [dit]
        else:
            self._filterItems[diag.key].insert(0, dit)
        return dit
        
    def reject(self):
        self.close()

    def close(self):
        self._parent.diagview = None
        QWidget.close(self)

# -----------------------------------------------------------------
